@@ -1,3 +1,5 @@
+/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
+
 #ifndef _WIN32
 #include <sys/ioctl.h>
 #include <termios.h>
@@ -37,8 +39,9 @@ struct cliopts_priv {
     int argsplit;
     int wanted;
 
-    char current_key[4096];
-    char current_value[4096];
+#define MAX_KEYLEN 4096
+    char current_key[MAX_KEYLEN];
+    char current_value[MAX_KEYLEN];
 };
 
 enum {
@@ -71,19 +74,25 @@ parse_option(struct cliopts_priv *ctx, const char *key);
 static int
 parse_value(struct cliopts_priv *ctx, const char *value);
 
+static void ensure_list_capacity(cliopts_list *l)
+{
+    if (l->nvalues == l->nalloc) {
+        if (l->nalloc == 0) {
+            l->nalloc = 2;
+            l->values = malloc(l->nalloc * sizeof(*l->values));
+        } else {
+            l->nalloc *= 1.5;
+            l->values = realloc(l->values, sizeof(*l->values) * l->nalloc);
+        }
+    }
+}
+
 static void
 add_list_value(const char *src, size_t nsrc, cliopts_list *l)
 {
     char *cp = malloc(nsrc + 1);
 
-    if (!l->nalloc) {
-        l->nalloc = 2;
-        l->values = malloc(l->nalloc * sizeof(*l->values));
-    } else {
-        l->nalloc *= 1.5;
-        l->values = realloc(l->values, sizeof(*l->values) * l->nalloc);
-    }
-
+    ensure_list_capacity(l);
     l->values[l->nvalues++] = cp;
     cp[nsrc] = '\0';
     memcpy(cp, src, nsrc);
@@ -103,6 +112,73 @@ cliopts_list_clear(cliopts_list *l)
     l->nalloc = 0;
 }
 
+static void ensure_pair_list_capacity(cliopts_pair_list *l)
+{
+    if (l->nvalues == l->nalloc) {
+        if (l->nalloc == 0) {
+            l->nalloc = 2;
+            l->keys = malloc(l->nalloc * sizeof(*l->keys));
+            l->values = malloc(l->nalloc * sizeof(*l->values));
+        } else {
+            l->nalloc *= 1.5;
+            l->keys = realloc(l->keys, sizeof(*l->keys) * l->nalloc);
+            l->values = realloc(l->values, sizeof(*l->values) * l->nalloc);
+        }
+    }
+}
+
+static void
+add_pair_list_value(const char *src, size_t nsrc, cliopts_pair_list *l)
+{
+    char *key = NULL, *val = NULL;
+    char *sep = memchr(src, '=', nsrc);
+    if (sep == NULL) {
+        key = malloc(nsrc + 1);
+        memcpy(key, src, nsrc);
+        key[nsrc] = '\0';
+        val = malloc(1);
+        val[0] = '\0';
+    } else {
+        char *pp = sep;
+        size_t nkey = sep - src;
+        size_t nval = nsrc - nkey - 1;
+        for (; pp > src; pp--, nkey--) {
+            if (*pp != ' ' && *pp != '\t' && *pp != '\0') {
+                break;
+            }
+        }
+        key = malloc(nkey + 1);
+        memcpy(key, src, nkey);
+        key[nkey] = '\0';
+        val = malloc(nval + 1);
+        memcpy(val, sep + 1, nval);
+        val[nval] = '\0';
+    }
+
+    ensure_pair_list_capacity(l);
+
+    l->keys[l->nvalues] = key;
+    l->values[l->nvalues] = val;
+    l->nvalues++;
+}
+
+CLIOPTS_API
+void
+cliopts_pair_list_clear(cliopts_pair_list *l)
+{
+    size_t ii;
+    for (ii = 0; ii < l->nvalues; ii++) {
+        free(l->keys[ii]);
+        free(l->values[ii]);
+    }
+    free(l->keys);
+    free(l->values);
+    l->keys = NULL;
+    l->values = NULL;
+    l->nvalues = 0;
+    l->nalloc = 0;
+}
+
 /**
  * Various extraction/conversion functions for numerics
  */
@@ -230,6 +306,7 @@ parse_value(struct cliopts_priv *ctx,
         char *vp = malloc(vlen+1);
         vp[vlen] = 0;
         strcpy(vp, value);
+        free(*(char**)entry->dest);
         *(char**)entry->dest = vp;
         return WANT_OPTION;
     }
@@ -238,6 +315,10 @@ parse_value(struct cliopts_priv *ctx,
         add_list_value(value, vlen, (cliopts_list *)entry->dest);
         return WANT_OPTION;
     }
+    if (entry->ktype == CLIOPTS_ARGT_PAIR_LIST) {
+        add_pair_list_value(value, vlen, (cliopts_pair_list *)entry->dest);
+        return WANT_OPTION;
+    }
 
     if (entry->ktype == CLIOPTS_ARGT_FLOAT) {
         exfn = extract_float;
@@ -304,6 +385,11 @@ parse_option(struct cliopts_priv *ctx,
         ctx->errnum = CLIOPTS_ERR_BADOPT;
         return MODE_ERROR;
     }
+    if (klen > MAX_KEYLEN) {
+        ctx->errstr = "The key is to big";
+        ctx->errnum = CLIOPTS_ERR_BADOPT;
+        return MODE_ERROR;
+    }
 
     /**
      * figure out what type of option it is..
@@ -327,6 +413,11 @@ parse_option(struct cliopts_priv *ctx,
             break;
         }
     }
+    if (valp && strlen(valp) > MAX_KEYLEN) {
+        ctx->errstr = "The value is to big";
+        ctx->errnum = CLIOPTS_ERR_BAD_VALUE;
+        return MODE_ERROR;
+    }
 
     GT_PARSEOPT:
     memset(ctx->current_value, 0, sizeof(ctx->current_value));
@@ -596,6 +687,17 @@ print_help(struct cliopts_priv *ctx, struct cliopts_extra_settings *settings)
                 }
                 break;
             }
+            case CLIOPTS_ARGT_PAIR_LIST: {
+                size_t ii;
+                cliopts_pair_list *l = (cliopts_pair_list *)cur->dest;
+                for (ii = 0; ii < l->nvalues; ii++) {
+                    fprintf(stderr, "'%s=%s'", l->keys[ii], l->values[ii]);
+                    if (ii != l->nvalues-1) {
+                        fprintf(stderr, ", ");
+                    }
+                }
+                break;
+            }
             case CLIOPTS_ARGT_FLOAT:
                 fprintf(stderr, "%0.2f", *(float*)cur->dest);
                 break;
@@ -723,7 +825,9 @@ cliopts_parse_options(cliopts_entry *entries,
             }
 
             print_help(&ctx, settings);
-            exit(0);
+            if (!settings->help_noexit) {
+                exit(0);
+            }
 
         } else if (curmode == MODE_RESTARGS) {
             ii++;
@@ -776,3 +880,85 @@ cliopts_parse_options(cliopts_entry *entries,
     }
     return ret;
 }
+
+CLIOPTS_API
+int cliopts_split_args(char *args, int *argc, char ***argv)
+{
+    char *p = args;
+    char *current = NULL;
+    int skip = 0;
+
+    *argc = 0;
+    *argv = NULL;
+
+#define ch(x) (*(x + skip))
+
+    while (1) {
+
+        while (ch(p) && isspace(ch(p))) {
+            p++;
+        }
+        if (*p) {
+            int insq = 0;
+            int done = 0;
+
+            if (current == NULL) {
+                current = p;
+            }
+            while (!done) {
+                if (insq) {
+                    if (ch(p) == '\\' && ch(p + 1) == '\'') {
+                        skip++;
+                    } else if (ch(p) == '\'') {
+                        if (ch(p + 1) && !isspace(ch(p + 1))) {
+                            // do not consider single quote inside word as terminating
+                        } else {
+                            *p = '\0';
+                            p++;
+                            done = 1;
+                        }
+                    } else if (!*p) {
+                        cliopt_debug("unterminated single quote");
+                        goto err;
+                    }
+                } else {
+                    switch (ch(p)) {
+                    case '\'':
+                        if (current == p) {
+                            current++;
+                            insq = 1;
+                        }
+                        break;
+                    case ' ':
+                    case '\n':
+                    case '\r':
+                    case '\t':
+                        *p = '\0';
+                        p++;
+                    case '\0':
+                        done = 1;
+                        break;
+                    }
+                }
+                if (skip) {
+                    *p = ch(p);
+                }
+                if (*p && !done) {
+                    p++;
+                }
+            }
+            *argv = realloc(*argv, (*argc + 1) * sizeof(char *));
+            (*argv)[*argc] = current;
+            (*argc)++;
+            current = NULL;
+        } else {
+            return 0;
+        }
+    }
+
+err:
+    free(*argv);
+    *argc = 0;
+    *argv = NULL;
+    return 1;
+}
