@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 /*
- *     Copyright 2013 Couchbase, Inc.
+ *     Copyright 2017-2020 Couchbase, Inc.
  *
  *   Licensed under the Apache License, Version 2.0 (the "License");
  *   you may not use this file except in compliance with the License.
@@ -15,7 +15,6 @@
  *   limitations under the License.
  */
 
-
 /**
  * @file
  * @brief Command codes for libcouchbase.
@@ -34,7 +33,6 @@
 extern "C" {
 #endif
 
-
 /**
  * @ingroup lcb-cntl
  * @defgroup lcb-cntl-settings Setting List
@@ -47,7 +45,7 @@ extern "C" {
  *
  * @code{.c}
  * char something;
- * lcb_error_t rv;
+ * lcb_STATUS rv;
  * rv = lcb_cntl(instance, LCB_CNTL_GET, LCB_CNTL_FOO, &something);
  * @endcode
  *
@@ -73,14 +71,25 @@ extern "C" {
  * pointer types depending on whether the `mode` is retrieval or storage.
  *
  *
- * @section lcb-timeout-info Timeout Settings
+ * @section lcb-time-info Timeout and Time Value Settings
+ *
+ * There are various settings on the library that control behavior with
+ * respect to wall clock time.
  *
  * Timeout settings control how long the library will wait for a certain event
  * before proceeding to the next course of action (which may either be to try
  * a different operation or fail the current one, depending on the specific
  * timeout).
  *
- * Timeouts are specified in _microseconds_ stored within an `lcb_U32`.
+ * Other settings may configure how often the library proactively polls for
+ * a configuration update, retries various interally retried operations and
+ * so forth.
+ *
+ * Time values are specified in _microseconds_ stored within an `lcb_U32`.
+ *
+ * When specified as an argument to lcb_cntl_string() or through the connection
+ * string, it will be parsed from a string float value where the integer-part
+ * is in seconds and the fractional-part is in fractions of a second.
  *
  * Note that timeouts in libcouchbase are implemented via an event loop
  * scheduler. As such their accuracy and promptness is limited by how
@@ -133,9 +142,9 @@ extern "C" {
  *
  * @cntl_arg_both{lcbU32*}
  * @committed
- * @see lcb-timeout-info
+ * @see lcb-time-info
  */
-#define LCB_CNTL_OP_TIMEOUT             0x00
+#define LCB_CNTL_OP_TIMEOUT 0x00
 
 /**
  * @brief Views Timeout
@@ -144,7 +153,7 @@ extern "C" {
  * @cntl_arg_both{lcb_U32*}
  * @committed
  */
-#define LCB_CNTL_VIEW_TIMEOUT           0x01
+#define LCB_CNTL_VIEW_TIMEOUT 0x01
 
 /**
  * @brief N1QL Timeout
@@ -153,7 +162,7 @@ extern "C" {
  * @cntl_arg_both{lcb_U32*}
  * @committed
  */
-#define LCB_CNTL_N1QL_TIMEOUT 0x3D
+#define LCB_CNTL_QUERY_TIMEOUT 0x3D
 
 /**
  * @brief Get the name of the bucket
@@ -163,7 +172,22 @@ extern "C" {
  * @cntl_arg_getonly{`const char*`}
  * @committed
  */
-#define LCB_CNTL_BUCKETNAME             0x30
+#define LCB_CNTL_BUCKETNAME 0x30
+
+/**
+ * @brief Get the bucket type.
+ * This returns the bucket type - which is either of the following:
+ *
+ *  * LCB_BTYPE_UNSPEC
+ *  * LCB_BTYPE_COUCHBASE
+ *  * LCB_BTYPE_EPHEMERAL
+ *  * LCB_BTYPE_MEMCACHED
+ *
+ * @see https://developer.couchbase.com/documentation/server/current/architecture/core-data-access-buckets.html
+ *
+ * @cntl_arg_getonly{lcb_BTYPE*}
+ */
+#define LCB_CNTL_BUCKETTYPE 0x48
 
 /**
  * @brief Get the handle type.
@@ -171,25 +195,22 @@ extern "C" {
  * LCB_TYPE_BUCKET
  *
  * @cntl_arg_getonly{lcb_type_t*}
- * @uncommitted
  */
-#define LCB_CNTL_HANDLETYPE             0x04
+#define LCB_CNTL_HANDLETYPE 0x04
 
 /**@brief Get the vBucket handle.
  * Obtains the current cluster configuration from the client.
  *
  * @cntl_arg_getonly{lcbvb_CONFIG**}
- * @uncommitted
  */
-#define LCB_CNTL_VBCONFIG               0x05
+#define LCB_CNTL_VBCONFIG 0x05
 
 /**@brief Get the iops implementation instance
  *
  * @cntl_arg_getonly{lcb_io_opt_t*}
  * @uncommitted
  */
-#define LCB_CNTL_IOPS                   0x06
-
+#define LCB_CNTL_IOPS 0x06
 
 /** @brief Structure containing mapping information for a key */
 typedef struct lcb_cntl_vbinfo_st {
@@ -198,9 +219,9 @@ typedef struct lcb_cntl_vbinfo_st {
     union {
         /** v0 */
         struct {
-            const void *key; /**< **Input** Key */
-            lcb_SIZE nkey; /**< **Input** Length of key */
-            int vbucket; /**< **Output** Mapped vBucket */
+            const void *key;  /**< **Input** Key */
+            lcb_SIZE nkey;    /**< **Input** Length of key */
+            int vbucket;      /**< **Output** Mapped vBucket */
             int server_index; /**< **Output** Server index for vBucket */
         } v0;
     } v;
@@ -212,7 +233,28 @@ typedef struct lcb_cntl_vbinfo_st {
  * @cntl_arg_getonly{lcb_cntl_vbinfo_t*}
  * @committed
  */
-#define LCB_CNTL_VBMAP                  0x07
+#define LCB_CNTL_VBMAP 0x07
+
+/**
+ * Modes for handling IPv6 in the IO layer.
+ */
+typedef enum {
+    LCB_IPV6_DISABLED = 0x00, /**< disable IPv6 */
+    LCB_IPV6_ONLY = 0x1,      /**< enforce only IPv6 */
+    LCB_IPV6_ALLOW = 0x02     /**< use both IPv6 and IPv4 */
+} lcb_ipv6_t;
+
+/**
+ * @brief IPv4/IPv6 selection policy
+ *
+ * Setting which controls whether hostname lookups should prefer IPv4 or IPv6
+ *
+ * Use `ipv6` in the connection string (e.g. "ipv6=allow" or "ipv6=only")
+ *
+ * @cntl_arg_both{lcb_ipv6_t*}
+ * @committed
+ */
+#define LCB_CNTL_IP6POLICY 0x0b
 
 /**
  * @brief Configuration error threshold.
@@ -222,13 +264,12 @@ typedef struct lcb_cntl_vbinfo_st {
  * update is requested again.
  *
  * @cntl_arg_both{lcb_SIZE*}
- * @uncommitted
  */
-#define LCB_CNTL_CONFERRTHRESH          0x0c
+#define LCB_CNTL_CONFERRTHRESH 0x0c
 
 /**
  * @brief Default timeout for lcb_durability_poll()
- * @ingroup lcb-timeout-info
+ * @ingroup lcb-time-info
  *
  * This is the time the client will
  * spend sending repeated probes to a given key's vBucket masters and replicas
@@ -237,21 +278,29 @@ typedef struct lcb_cntl_vbinfo_st {
  * @cntl_arg_both{lcb_U32*}
  * @committed
  */
-#define LCB_CNTL_DURABILITY_TIMEOUT     0x0d
+#define LCB_CNTL_DURABILITY_TIMEOUT 0x0d
 
-/**@brief Polling grace interval for lcb_durability_poll()
+/**
+ * @brief Polling grace interval for lcb_durability_poll()
  *
  * This is the time the client will wait between repeated probes to
  * a given server.
  *
  * @cntl_arg_both{lcb_U32*}
- * @committed*/
-#define LCB_CNTL_DURABILITY_INTERVAL    0x0e
+ * @committed
+ */
+#define LCB_CNTL_DURABILITY_INTERVAL 0x0e
 
-/**@brief Timeout for non-views HTTP requests
+/**
+ * @brief Timeout for otherwise unspecified HTTP requests
+ *
+ * Examples of these kinds of HTTP requests might be cluster management,
+ * user management, etc.
+ *
  * @cntl_arg_both{lcb_U32*}
- * @committed*/
-#define LCB_CNTL_HTTP_TIMEOUT           0x0f
+ * @committed
+ */
+#define LCB_CNTL_HTTP_TIMEOUT 0x0f
 
 /**
  * @brief Print verbose plugin load information to console
@@ -268,14 +317,14 @@ typedef struct lcb_cntl_vbinfo_st {
  * @note Pass NULL to lcb_cntl for the 'instance' parameter.
  * @volatile
  */
-#define LCB_CNTL_IOPS_DLOPEN_DEBUG       0x11
+#define LCB_CNTL_IOPS_DLOPEN_DEBUG 0x11
 
 /**@brief Initial bootstrap timeout.
  * This is how long the client will wait to obtain the initial configuration.
  *
  * @cntl_arg_both{lcb_U32*}
  * @committed*/
-#define LCB_CNTL_CONFIGURATION_TIMEOUT   0x12
+#define LCB_CNTL_CONFIGURATION_TIMEOUT 0x12
 
 /**
  * @brief Randomize order of bootstrap nodes.
@@ -323,7 +372,6 @@ typedef struct lcb_cntl_vbinfo_st {
  * libcouchbase API, after which it may have been freed.
  *
  * @cntl_arg_get_and_set{char**, char*}
- * @uncommitted
  */
 #define LCB_CNTL_FORCE_SASL_MECH 0x16
 
@@ -350,47 +398,8 @@ typedef struct lcb_cntl_vbinfo_st {
  * @{
  */
 
-/** @brief Logging Levels */
-typedef enum { LCB_LOG_TRACE = 0, LCB_LOG_DEBUG, LCB_LOG_INFO, LCB_LOG_WARN,
-    LCB_LOG_ERROR, LCB_LOG_FATAL, LCB_LOG_MAX
-} lcb_log_severity_t;
-
-struct lcb_logprocs_st;
-
 /**
- * @brief Logger callback
- *
- * @uncommitted
- *
- * This callback is invoked for each logging message emitted
- * @param procs the logging structure provided
- * @param iid instance id
- * @param subsys a string describing the module which emitted the message
- * @param severity one of the LCB_LOG_* severity constants.
- * @param srcfile the source file which emitted this message
- * @param srcline the line of the file for the message
- * @param fmt a printf format string
- * @param ap a va_list for vprintf
- */
-typedef void (*lcb_logging_callback)(struct lcb_logprocs_st *procs,
-        unsigned int iid, const char *subsys, int severity, const char *srcfile,
-        int srcline, const char *fmt, va_list ap);
-
-/**
- * @brief Logging context
- * @uncommitted
- *
- * This structure defines the logging handlers. Currently there is only
- * a single field defined which is the default callback for the loggers.
- * This API may change.
- */
-typedef struct lcb_logprocs_st {
-    int version;
-    union { struct { lcb_logging_callback callback; } v0; } v;
-} lcb_logprocs;
-
-/**
- * @brief Access the lcb_logprocs structure
+ * @brief Access the lcb_LOGGER structure
  * @uncommitted
  *
  * The lcb_logoprocs structure passed must not be freed until the instance
@@ -399,8 +408,8 @@ typedef struct lcb_logprocs_st {
  *
  * @cntl_arg_get_and_set{lcb_logprocs**,lcb_logprocs*}*/
 #define LCB_CNTL_LOGGER 0x18
-/**@}*/
 
+/**@}*/
 
 /**
  * @brief Refresh Throttling
@@ -415,13 +424,14 @@ typedef struct lcb_logprocs_st {
  *
  * @cntl_arg_both{lcb_U32*}
  * @see LCB_CNTL_CONFERRTHRESH
- * @uncommitted
  */
 #define LCB_CNTL_CONFDELAY_THRESH 0x19
 
-/**@brief Get the transport used to fetch cluster configuration.
+/**
+ * @brief Get the transport used to fetch cluster configuration.
  * @cntl_arg_getonly{lcb_config_transport_t*}
- * @uncommitted*/
+ * @uncommitted
+ */
 #define LCB_CNTL_CONFIG_TRANSPORT 0x1A
 
 /**
@@ -468,19 +478,22 @@ typedef struct lcb_logprocs_st {
  */
 #define LCB_CNTL_HTCONFIG_IDLE_TIMEOUT 0x1C
 
-/**@brief Get the current SCM changeset for the library binary
+/**
+ * @brief Get the current SCM changeset for the library binary
  * @cntl_arg_getonly{char**}
- * @committed*/
+ */
 #define LCB_CNTL_CHANGESET 0x1F
 
 /**
- * @brief file used for the configuration cache.
+ * @brief File used for the configuration cache.
  *
  * The configuration
  * cache allows to bootstrap from a cluster without using the initial
  * bootstrap connection, considerably reducing latency. If the file passed
  * does not exist, the normal bootstrap process is performed and the file
- * is written to with the current information.
+ * is written to with the current information.  File will be updated as
+ * the configuration in the cluster changes.  Multiple instances may race
+ * to update the file, and that is the intended behavior.
  *
  * @note The leading directories for the file must exist, otherwise the file
  * will never be created.
@@ -503,9 +516,14 @@ typedef struct lcb_logprocs_st {
  */
 #define LCB_CNTL_CONFIGCACHE_RO 0x36
 
+/**
+ * SSL options
+ *
+ * @committed
+ */
 typedef enum {
-    LCB_SSL_ENABLED = 1 << 0, /**< Use SSL */
-    LCB_SSL_NOVERIFY = 1 << 1, /**< Don't verify certificates */
+    LCB_SSL_ENABLED = 1 << 0,     /**< Use SSL */
+    LCB_SSL_NOVERIFY = 1 << 1,    /**< Don't verify certificates */
     LCB_SSL_NOGLOBALINIT = 1 << 2 /**< Do not call SSL's global init functions */
 } lcb_SSLOPTS;
 
@@ -517,6 +535,7 @@ typedef enum {
  * within the connection string. See @ref lcb_create_st3 for details.
  *
  * @cntl_arg_getonly{`int*` (value is one of @ref lcb_SSLOPTS)}
+ * @committed
  */
 #define LCB_CNTL_SSL_MODE 0x22
 
@@ -527,19 +546,49 @@ typedef enum {
  *
  * @cntl_arg_getonly{`char**`}
  * @see LCB_CNTL_SSL_MODE
+ * @committed
  */
 #define LCB_CNTL_SSL_CERT 0x23
-/* For back compat */
+
+/**
+ * @brief Get SSL private key path
+ *
+ * Retrieve the path to the private key (if any) being used.
+ * When key specified, the library will use it to authenticate on the services,
+ * skipping all other authentication mechanisms (SASL, HTTP Basic auth, etc)
+ *
+ * @cntl_arg_getonly{`char**`}
+ * @see LCB_CNTL_SSL_MODE
+ * @see https://developer.couchbase.com/documentation/server/5.0/security/security-certs-auth.html
+ * @committed
+ */
+#define LCB_CNTL_SSL_KEY 0x4b
+
+/**
+ * @brief Get SSL trust store path
+ *
+ * Trust store might be NULL, in this case the library expects it to be concatenated with certificate.
+ *
+ * @cntl_arg_getonly{`char**`}
+ * @see LCB_CNTL_SSL_MODE
+ * @see https://developer.couchbase.com/documentation/server/5.0/security/security-certs-auth.html
+ * @committed
+ */
+#define LCB_CNTL_SSL_TRUSTSTORE 0x4d
+
+/**
+ * Alias for @ref LCB_CNTL_SSL_CERT for backward compatibility.
+ * @deprecated
+ */
 #define LCB_CNTL_SSL_CACERT LCB_CNTL_SSL_CERT
 
 /**
- * @brief
- * Select retry mode to manipulate
+ * @brief Select retry mode to manipulate
  */
 typedef enum {
     LCB_RETRY_ON_TOPOCHANGE = 0, /**< Select retry for topology */
-    LCB_RETRY_ON_SOCKERR, /**< Select retry for network errors */
-    LCB_RETRY_ON_VBMAPERR, /**< Select retry for NOT_MY_VBUCKET responses */
+    LCB_RETRY_ON_SOCKERR,        /**< Select retry for network errors */
+    LCB_RETRY_ON_VBMAPERR,       /**< Select retry for NOT_MY_VBUCKET responses */
 
     /** Retry when there is no node for the item. This case is special as the
      * `cmd` setting is treated as a boolean rather than a bitmask*/
@@ -577,8 +626,10 @@ typedef enum {
  */
 #define LCB_RETRYOPT_CREATE(mode, policy) (((mode) << 16) | policy)
 
-#define LCB_RETRYOPT_GETMODE(u) (u) >> 16
-#define LCB_RETRYOPT_GETPOLICY(u) (u) & 0xffff
+/** Get mode from retry setting value */
+#define LCB_RETRYOPT_GETMODE(u) ((u) >> 16)
+/** Get policy from retry setting value */
+#define LCB_RETRYOPT_GETPOLICY(u) ((u)&0xffff)
 
 /**
  * @volatile
@@ -600,7 +651,7 @@ typedef enum {
  * @code{.c}
  * for (int ii = 0; ii < LCB_RETRY_ON_MAX; ++ii) {
  *   lcb_U32 val = LCB_RETRYOPT_CREATE(ii, LCB_RETRY_CMDS_NONE);
- *   lcb_error_t err = lcb_cntl(instance, LCB_CNTL_SET, LCB_CNTL_RETRYMODE, &val);
+ *   lcb_STATUS err = lcb_cntl(instance, LCB_CNTL_SET, LCB_CNTL_RETRYMODE, &val);
  * }
  * @endcode
  *
@@ -637,12 +688,9 @@ typedef enum {
 } lcb_HTCONFIG_URLTYPE;
 
 /**
- * @volatile - Primarily here to support tests and buggy HTTP servers/proxies
- * which do not like to maintain a connection upon receipt of a 404.
- *
  * @brief Set the URL selection mode.
  *
- * The URL type can be a mask of the lcb_HTCONFIG_URLTYPE constants which
+ * The URL type can be a mask of the #lcb_HTCONFIG_URLTYPE constants which
  * indicate which URLs the HTTP provider should use.
  *
  * The default is to use the `25PLUS` URI first, and fallback on the compat uri
@@ -654,11 +702,13 @@ typedef enum {
  * This setting is only used when CCCP is disabled. This will typically be for
  * older clusters or for memcached buckets.
  * @cntl_arg_both{`int*` (value is one of @ref lcb_HTCONFIG_URLTYPE)}
+
+ * @volatile Primarily here to support tests and buggy HTTP servers/proxies
+ * which do not like to maintain a connection upon receipt of a 404.
  */
 #define LCB_CNTL_HTCONFIG_URLTYPE 0x25
 
 /**
- * @volatile
  * Determines whether to run the event loop internally within lcb_destroy()
  * until no more I/O resources remain for the library. This is usually only
  * necessary if you are creating a lot of instances and/or are using memory
@@ -666,12 +716,11 @@ typedef enum {
  *
  * @cntl_arg_both{`int*` (as a boolean)}
  * @see lcb_destroy_async() and lcb_set_destroy_callback()
+ * @volatile
  */
 #define LCB_CNTL_SYNCDESTROY 0x28
 
 /**
- * @committed
- *
  * Sets the logging level for the console logger. If a logger is already
  * initialized (either from the environment, or via lcb_cntl_logger() then
  * this operation does nothing.
@@ -681,8 +730,8 @@ typedef enum {
  * it from the environment.
  *
  * The argument passed to lcb_cntl() is an integer of 0 until
- * `LCB_LOGLEVEL_MAX`, though the actual type is of `lcb_U32` rather than
- * an enum type.
+ * `LCB_LOG_MAX`, though the actual type is of `lcb_U32` rather than
+ * an enum type #lcb_log_severity_t.
  *
  * @cntl_arg_setonly{const lcb_U32 *}
  * @see LCB_CNTL_LOGGER
@@ -690,7 +739,6 @@ typedef enum {
 #define LCB_CNTL_CONLOGGER_LEVEL 0x29
 
 /**
- * @uncommitted
  *
  * Sets the output file (as a `FILE*`) for the console logger. Note that
  * any existing file pointer will be cleared (but not `fclose()`d.
@@ -710,10 +758,8 @@ typedef enum {
 #define LCB_CNTL_CONLOGGER_FP 0x3B
 
 /**
- * @committed
- *
  * Sets the behavior for reporting network errors. By default network errors
- * are returned as `LCB_NETWORK_ERROR` return codes for compatibility reasons.
+ * are returned as `LCB_ERR_NETWORK` return codes for compatibility reasons.
  * More detailed error codes may be available by enabling this option which will
  * return appropriate error codes which have a category of
  * @ref LCB_ERRTYPE_NETWORK
@@ -726,7 +772,6 @@ typedef enum {
 #define LCB_CNTL_DETAILED_ERRCODES 0x2A
 
 /**
- * @uncommitted
  *
  * Sets the interval at which the retry queue will attempt to resend a failed
  * operation. When an operation fails and the retry policy (see
@@ -738,50 +783,35 @@ typedef enum {
  * network traffic for operations which are not yet ready to be retried.
  *
  * @cntl_arg_both{lcb_U32* (microseconds)}
- *
- * @see LCB_CNTL_RETRY_BACKOFF
  */
 #define LCB_CNTL_RETRY_INTERVAL 0x2C
 
 /**
- * @uncommitted
- *
- * When an operation has been retried more than once and it has still not
- * succeeded, the library will attempt to back off for the operation by
- * scheduling it to be retried in `LCB_CNTL_RETRY_INTEVAL * ${n}` microseconds,
- * where `${n}` is the factor controlled by this setting.
- *
- * @cntl_arg_both{float*}
- */
-#define LCB_CNTL_RETRY_BACKOFF 0x2D
-
-/**
- * @volatile
  * Whether commands are retried immediately upon receipt of not-my-vbucket
  * replies.
  *
  * Since version 2.4.8, packets by default are retried immediately on a
  * different node if it had previously failed with a not-my-vbucket
  * response, and is thus not subject to the @ref LCB_CNTL_RETRY_INTERVAL
- * and @ref LCB_CNTL_RETRY_BACKOFF settings. Disabling this setting will
- * restore the older behavior. This may be used in case there are problems
- * with the default heuristic/retry algorithm.
+ * setting. Disabling this setting will restore the older behavior.
+ * This may be used in case there are problems with the default
+ * heuristic/retry algorithm.
+ *
+ * @volatile
  */
 #define LCB_CNTL_RETRY_NMV_IMM 0x37
 
 /**
- * @volatile
- *
  * Set the maximum pool size for pooled http (view request) sockets. This should
  * be set to 1 (the default) unless you plan to execute concurrent view requests.
  * You may set this to 0 to disable pooling
  *
  * @cntl_arg_both{lcb_SIZE}
+ * @volatile
  */
 #define LCB_CNTL_HTTP_POOLSIZE 0x2E
 
 /**
- * @uncomitted
  * Determine whether or not a new configuration should be received when an error
  * is received over the HTTP API (i.e. via lcb_make_http_request().
  *
@@ -789,24 +819,24 @@ typedef enum {
  * expectedly issuing a lot of requests which may result in an error.
  *
  * @cntl_arg_both{int (as boolean)}
+ * @uncommitted
  */
 #define LCB_CNTL_HTTP_REFRESH_CONFIG_ON_ERROR 0x2F
 
 /**
- * @volatile
  * Set the behavior of the lcb_sched_leave() API call. By default the
  * lcb_sched_leave() will also set up the necessary requirements for flushing
  * to the network. If this option is off then an explicit call to
  * lcb_sched_flush() must be performed instead.
  *
  * @cntl_arg_both{int (as boolean)}
+ * @volatile
  */
 #define LCB_CNTL_SCHED_IMPLICIT_FLUSH 0x31
 
 /**
- * @volatile
  *
- * Allow the server to return an additional 16 bytes of data for each
+ * Request the server to return an additional 16 bytes of data for each
  * mutation operation. This extra information may help with more reliable
  * durability polling, but will also increase the size of the response packet.
  *
@@ -816,45 +846,21 @@ typedef enum {
  *
  * @cntl_arg_both{int (as boolean)}
  */
-#define LCB_CNTL_FETCH_MUTATION_TOKENS 0x34
+#define LCB_CNTL_ENABLE_MUTATION_TOKENS 0x34
 
 /**
- * @volatile
- *
- * This setting determines whether the lcb_durability_poll() function will
- * transparently attempt to use mutation token functionality (rather than checking
- * the CAS). This option is most useful for older code which does
- * explicitly use mutation tokens but would like to use its benefits when
- * ensuring durability constraints are satisfied.
- *
- * This option is enabled by default. Users may wish to disable this if they
- * are performing durability operations against items stored from different
- * client instances, as this will make use of a client-global state which is
- * derived on a per-vBucket basis. This means that the last mutation performed
- * on a given vBucket for the client will be used, which in some cases may be
- * older or newer than the mutations passed to the lcb_durability_poll()
- * function.
- *
- * @cntl_arg_both{int (as boolean)}
- */
-#define LCB_CNTL_DURABILITY_MUTATION_TOKENS 0x35
-
-/**
- * @volatile
- *
  * This read-only property determines if the mutation token mechanism is supported
  * on the cluster itself. This will only be accurate once a single operation
  * has been performed on the cluster - or in other words, once a connection
  * to a data node has been established for the purposes of normal operations.
  *
  * @cntl_arg_getonly{int (as boolean)}
+ * @uncommitted
  */
 #define LCB_CNTL_MUTATION_TOKENS_SUPPORTED 0x38
 
-
 /**
- * @uncommitted
- * This setting determines if calls to lcb_wait() and lcb_wait3() will reset
+ * This setting determines if calls to lcb_wait() and lcb_wait() will reset
  * the timeout of pending operations to the time that lcb_wait() was called,
  * rather than having the operation maintain the time of the call which
  * scheduled it. If the time between lcb_store3() and family and the lcb_wait()
@@ -862,21 +868,21 @@ typedef enum {
  * avoid prematurely having operations time out.
  *
  * @cntl_arg_both{int (as boolean)}
+ * @uncommitted
  *
  * Use `"readj_wait_tmo"` for the string version
  */
 #define LCB_CNTL_RESET_TIMEOUT_ON_WAIT 0x3A
 
 /**
- * @volatile
  * Clears the internal prepared statement cache for N1QL
  *
  * This does not take any arguments, and is valid only on @ref LCB_CNTL_SET
+ * @uncommitted
  */
-#define LCB_CNTL_N1QL_CLEARACHE 0x3E
+#define LCB_CNTL_QUERY_CLEARACHE 0x3E
 
 /**
- * @committed
  * Sets additional text for negotiation. This allows wrappers or applications
  * to add additional identifying information which can then be seen in the
  * server logs.
@@ -902,12 +908,11 @@ typedef const char *lcb_BUCKETCRED[2];
 /**
  * Set the amount of time the client should wait before retrying a
  * not-my-vbucket response packet. The default is 100ms. The value should
- * be specified in microseconds
- *
- * @committed
- * @cntl_arg_both{lcb_U32*}
+ * be specified in microseconds.
  *
  * Use `"retry_nmv_interval"` with lcb_cntl_string()
+ *
+ * @cntl_arg_both{lcb_U32*}
  */
 #define LCB_CNTL_RETRY_NMV_INTERVAL 0x41
 
@@ -919,13 +924,384 @@ typedef const char *lcb_BUCKETCRED[2];
  * @note This setting only works for event-style I/O plugins. This means it
  * has no effect on completion style plugins such as libuv or Windows IOCP
  *
- * @committed
  * @cntl_arg_both{lcb_U32*}
  */
 #define LCB_CNTL_READ_CHUNKSIZE 0x42
 
-/** This is not a command, but rather an indicator of the last item */
-#define LCB_CNTL__MAX                    0x43
+/**
+ * Enable/Disable sending the SELECT_BUCKET command after authentication.
+ * This is useful to test auth, and should not be set by end-users.
+ *
+ * Note that even if this feature is enabled (the default), the client will
+ * only send `SELECT_BUCKET` if the server indicates that it is supported
+ * during negotiation.
+ *
+ * Use `select_bucket` in the connection string
+ *
+ * @cntl_arg_both{int* (as boolean)}
+ */
+#define LCB_CNTL_SELECT_BUCKET 0x44
+
+/**
+ * Enable/Disable setting the `TCP_KEEPALIVE` option on created sockets.
+ * This is enabled by default for I/O backends which support it.
+ *
+ * The keepalive interval will be set to the operating system default.
+ *
+ * @cntl_arg_both{int* (as boolean)}
+ */
+#define LCB_CNTL_TCP_KEEPALIVE 0x45
+
+/**
+ * Set the amount of time to wait in between polling for a new configuration.
+ * This will have no effect if connected to a Memcached buckets, or using
+ * HTTP or File-based configurations (see the `bootstrap_on` connection
+ * string option).
+ *
+ * This option facilitates 'fast failover' - in that the client can preemptively
+ * check for any cluster topology updates before encountering an error.
+ *
+ * @cntl_arg_both{lcb_U32*}
+ *
+ * The value for this option is a time value. See the top of this header
+ * in respect to how to specify this.
+ *
+ * Using a value of `0` disables this feature.
+ *
+ * You can also use `config_poll_interval` in the connection string.
+ *
+ * @note
+ * Background polling is implemented in the library's non-blocking event loop.
+ * Synchronous clients (i.e. those using `lcb_wait()`) will only be able to
+ * poll as often as the library's event loop is active. If the library is
+ * suspended, that is, if not inside an `lcb_wait()` call, the library will
+ * be unable to do any kind of background polling.
+ */
+#define LCB_CNTL_CONFIG_POLL_INTERVAL 0x46
+
+/**
+ * Once redaction is enabled, anything at ERROR, WARN and INFO will wrap
+ * sensitive information with special tags, for further processing with the goal
+ * to remove or encrypt that information.  DEBUG or TRACE level logging are
+ * expected to have specific info.
+ *
+ * Use `log_redaction` in the connection string
+ *
+ * @cntl_arg_both{int* (as boolean)}
+ * @committed
+ */
+#define LCB_CNTL_LOG_REDACTION 0x4c
+
+/**
+ * Activate/deactivate end-to-end tracing.
+ *
+ * Use `enable_tracing` in the connection string
+ *
+ * @cntl_arg_both{int* (as boolean)}
+ * @see lcb-tracing-api
+ * @committed
+ */
+#define LCB_CNTL_ENABLE_TRACING 0x4e
+
+typedef enum {
+    LCBTRACE_THRESHOLD_KV = 0,
+    LCBTRACE_THRESHOLD_QUERY,
+    LCBTRACE_THRESHOLD_VIEW,
+    LCBTRACE_THRESHOLD_SEARCH,
+    LCBTRACE_THRESHOLD_ANALYTICS,
+    LCBTRACE_THRESHOLD__MAX
+} lcbtrace_THRESHOLDOPTS;
+
+/**
+ * Flush interval for orphaned spans queue in default tracer.
+ *
+ * This is the time the tracer will wait between repeated attempts
+ * to flush most recent orphaned spans.
+ *
+ * Use `tracing_orphaned_queue_flush_interval` in the connection string
+ *
+ * @code{.c}
+ * lcb_U32 tmo = 10000000; // 10 seconds in microseconds
+ * lcb_cntl(instance, LCB_CNTL_SET, LCB_CNTL_TRACING_ORPHANED_QUEUE_FLUSH_INTERVAL, &tmo);
+ * @endcode
+ *
+ * @code{.c}
+ * rv = lcb_cntl_string("tracing_orphaned_queue_flush_interval", "10.0");
+ * @endcode
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_TRACING_ORPHANED_QUEUE_FLUSH_INTERVAL 0x4f
+
+/**
+ * Size of orphaned spans queue in default tracer.
+ *
+ * Queues in default tracer has fixed size, and it will remove information about older spans,
+ * when the limit will be reached before flushing time.
+ *
+ * Use `tracing_orphaned_queue_size` in the connection string
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_TRACING_ORPHANED_QUEUE_SIZE 0x50
+
+/**
+ * Flush interval for spans with total time over threshold in default tracer.
+ *
+ * This is the time the tracer will wait between repeated attempts
+ * to flush threshold queue.
+ *
+ * Use `tracing_threshold_queue_flush_interval` in the connection string
+ *
+ * @code{.c}
+ * lcb_U32 tmo = 10000000; // 10 seconds in microseconds
+ * lcb_cntl(instance, LCB_CNTL_SET, LCB_CNTL_TRACING_THRESHOLD_QUEUE_FLUSH_INTERVAL, &tmo);
+ * @endcode
+ *
+ * @code{.c}
+ * rv = lcb_cntl_string("tracing_threshold_queue_flush_interval", "10.0");
+ * @endcode
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_TRACING_THRESHOLD_QUEUE_FLUSH_INTERVAL 0x51
+
+/**
+ * Size of threshold queue in default tracer.
+ *
+ * Queues in default tracer has fixed size, and it will remove information about older spans,
+ * when the limit will be reached before flushing time.
+ *
+ * Use `tracing_threshold_queue_size` in the connection string
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_TRACING_THRESHOLD_QUEUE_SIZE 0x52
+
+/**
+ * Minimum time for the tracing span of KV service to be considered by threshold tracer.
+ *
+ * Use `tracing_threshold_kv` in the connection string
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_TRACING_THRESHOLD_KV 0x53
+
+/**
+ * Minimum time for the tracing span of N1QL service to be considered by threshold tracer.
+ *
+ * Use `tracing_threshold_query` in the connection string
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_TRACING_THRESHOLD_QUERY 0x54
+
+/**
+ * Minimum time for the tracing span of VIEW service to be considered by threshold tracer.
+ *
+ * Use `tracing_threshold_view` in the connection string
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_TRACING_THRESHOLD_VIEW 0x55
+
+/**
+ * Minimum time for the tracing span of FTS service to be considered by threshold tracer.
+ *
+ * Use `tracing_threshold_search` in the connection string
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_TRACING_THRESHOLD_SEARCH 0x56
+
+/**
+ * Minimum time for the tracing span of ANALYTICS service to be considered by threshold tracer.
+ *
+ * Use `tracing_threshold_analytics` in the connection string
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_TRACING_THRESHOLD_ANALYTICS 0x57
+
+/**
+ * Options for how to handle compression
+ *
+ * @committed
+ */
+typedef enum {
+    /** Do not perform compression in any direction. Data which is received
+     * compressed via the server will be indicated as such by having the
+     * `LCB_VALUE_F_SNAPPYCOMP` flag set in the lcb_GETRESPv0::datatype field */
+    LCB_COMPRESS_NONE = 0x00,
+
+    /**
+     * Decompress incoming data, if the data has been compressed at the server.
+     * If this is set, the `datatype` field in responses will always be stripped
+     * of the `LCB_VALUE_F_SNAPPYCOMP` flag.
+     */
+    LCB_COMPRESS_IN = 1 << 0,
+
+    /**
+     * Compress outgoing data. Note that if the `datatype` field contains the
+     * `LCB_VALUE_F_SNAPPYCOMP` flag, then the data will never be compressed
+     * as it is assumed that it is already compressed.
+     */
+    LCB_COMPRESS_OUT = 1 << 1,
+
+    LCB_COMPRESS_INOUT = (LCB_COMPRESS_IN | LCB_COMPRESS_OUT),
+
+    /**
+     * By default the library will send a HELLO command to the server to
+     * determine whether compression is supported or not. Because commands may
+     * be pipelined prior to the scheduing of the HELLO command it is possible
+     * that the first few commands may not be compressed when schedule due to
+     * the library not yet having negotiated settings with the server. Setting
+     * this flag will force the client to assume that all servers support
+     * compression despite a HELLO not having been intially negotiated.
+     */
+    LCB_COMPRESS_FORCE = 1 << 2
+} lcb_COMPRESSOPTS;
+
+/**
+ * @brief Control how the library handles compression and deflation to and from
+ * the server.
+ *
+ * Starting in Couchbase Server 3.0, compression can optionally be applied to
+ * incoming and outcoming data. For incoming (i.e. `GET` requests) the data
+ * may be received in compressed format and then allow the client to inflate
+ * the data upon receipt. For outgoing (i.e. `SET` requests) the data may be
+ * compressed on the client side and then be stored and recognized on the
+ * server itself.
+ *
+ * The default behavior is to transparently handle compression for both incoming
+ * and outgoing data.
+ *
+ * Note that if the lcb_STORECMDv0::datatype field is set with compression
+ * flags, the data will _never_ be compressed by the library as this is an
+ * indication that it is _already_ compressed.
+ *
+ * @cntl_arg_both{`int*` (value is one of @ref lcb_COMPRESSOPTS)}
+ * @committed
+ */
+#define LCB_CNTL_COMPRESSION_OPTS 0x26
+
+/**
+ * Minimum size of the document payload to be compressed when compression enabled.
+ *
+ * Use `compression_min_size` in the connection string
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_COMPRESSION_MIN_SIZE 0x58
+
+/**
+ * Minimum compression ratio (compressed / original) of the compressed payload to allow sending it to cluster.
+ *
+ * Use `compression_min_ratio` in the connection string
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_COMPRESSION_MIN_RATIO 0x59
+
+/**
+ * Select type of network (alternative addresses).
+ *
+ * Use `network` in the connection string
+ *
+ * @cntl_arg_get_and_set{`const char**`, `const char*`}
+ *
+ * @uncommitted
+ */
+#define LCB_CNTL_NETWORK 0x5b
+
+/**
+ * The amount of time the pool should wait before closing idle connections.
+ *
+ * Use `http_pool_timeout` in the connection string
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_HTTP_POOL_TIMEOUT 0x5d
+
+/**
+ *
+ * @cntl_arg_both{int (as boolean)}
+ * @volatile
+ */
+#define LCB_CNTL_ENABLE_COLLECTIONS 0x4a
+
+/**
+ *
+ * @cntl_arg_both{int (as boolean)}
+ * @volatile
+ */
+#define LCB_CNTL_ENABLE_DURABLE_WRITE 0x5e
+
+/**
+ *
+ * @volatile
+ */
+#define LCB_CNTL_PERSISTENCE_TIMEOUT_FLOOR 0x5f
+
+/**
+ *
+ * @volatile
+ */
+#define LCB_CNTL_ALLOW_STATIC_CONFIG 0x60
+
+/**
+ * @brief Analytics Timeout
+ * This is the global I/O timeout for Analytics queries, issued via lcb_analytics()
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_ANALYTICS_TIMEOUT 0x61
+
+/**
+ * Tell the server that the library supports reordering the execution of the commands
+ * @cntl_arg_both{int (as boolean)}
+ * @volatile
+ */
+#define LCB_CNTL_ENABLE_UNORDERED_EXECUTION 0x62
+
+/**
+ * @brief Search Timeout
+ * This is the global I/O timeout for Search queries, issued via lcb_search()
+ *
+ * @cntl_arg_both{lcb_U32*}
+ * @committed
+ */
+#define LCB_CNTL_SEARCH_TIMEOUT 0x63
+
+/**
+ * Enable/Disable the Error Map feature. This is disabled by default.
+ * Works only on servers which support error map
+ *
+ * Use `enable_errmap` in the connection string
+ *
+ * @cntl_arg_both{int* (as boolean)}
+ */
+
+#define LCB_CNTL_ENABLE_ERRMAP 0x65
+
+/**
+ * This is not a command, but rather an indicator of the last item.
+ * @internal
+ */
+#define LCB_CNTL__MAX 0x66
 /**@}*/
 
 #ifdef __cplusplus
