@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 /*
- *     Copyright 2013 Couchbase, Inc.
+ *     Copyright 2013-2020 Couchbase, Inc.
  *
  *   Licensed under the Apache License, Version 2.0 (the "License");
  *   you may not use this file except in compliance with the License.
@@ -20,6 +20,7 @@
 
 #include "hostlist.h"
 #include <list>
+#include <utility>
 #include <lcbio/timer-ng.h>
 #include <lcbio/timer-cxx.h>
 
@@ -87,8 +88,10 @@
  *@{
  */
 
-namespace lcb {
-namespace clconfig {
+namespace lcb
+{
+namespace clconfig
+{
 
 /**
  * @brief Enumeration of the various config providers available.
@@ -103,6 +106,8 @@ enum Method {
     CLCONFIG_HTTP,
     /** Raw memcached provided */
     CLCONFIG_MCRAW,
+    /** Cluster administration provider. Static config with services */
+    CLCONFIG_CLADMIN,
 
     CLCONFIG_MAX,
 
@@ -110,7 +115,6 @@ enum Method {
     CLCONFIG_PHONY
 };
 
-
 /** Event types propagated to listeners */
 enum EventType {
     /** Called when a new configuration is being set in confmon */
@@ -126,7 +130,6 @@ enum EventType {
     CLCONFIG_EVENT_MONITOR_STOPPED
 };
 
-
 /** @brief Possible confmon states */
 enum State {
     /** The monitor is idle and not requesting a new configuration */
@@ -139,7 +142,6 @@ enum State {
     CONFMON_S_ITERGRACE = 1 << 1
 };
 
-
 struct Provider;
 struct Listener;
 class ConfigInfo;
@@ -154,29 +156,33 @@ struct Confmon {
      * This function creates a new `confmon` object which can be used to manage
      * configurations and their providers.
      *
-     * @param settings
-     * @param iot
+     * @param settings pointer to LCB settings
+     * @param iot pointer socket IO routines
+     * @param instance LCB handle
      *
      * Once the confmon object has been created you may enable or disable various
      * providers (see lcb_confmon_set_provider_active()). Once no more providers
      * remain to be activated you should call lcb_confmon_prepare() once. Then
      * call the rest of the functions.
      */
-    Confmon(lcb_settings*, lcbio_pTABLE iot);
-    void destroy() { delete this; }
+    Confmon(lcb_settings *settings, lcbio_pTABLE iot, lcb_INSTANCE *instance);
+    void destroy()
+    {
+        delete this;
+    }
     ~Confmon();
 
     /**
-     * Get the provider following the current provider, or NULL if this is
+     * Get the provider following the current provider, or nullptr if this is
      * the last provider in the list.
      * @param cur The current provider.
-     * @return The next provider, or NULL if no more providers remain.
+     * @return The next provider, or nullptr if no more providers remain.
      */
     Provider *next_active(Provider *cur);
 
     /**
      * Gets the first active provider.
-     * @return the first provider, or NULL if no providers exist.
+     * @return the first provider, or nullptr if no providers exist.
      */
     Provider *first_active();
 
@@ -184,7 +190,6 @@ struct Confmon {
      * Prepares the configuration monitor object for operations. This will insert
      * all the enabled providers into a list. Call this function each time a
      * provider has been enabled.
-     * @param mon
      */
     void prepare();
 
@@ -210,12 +215,11 @@ struct Confmon {
      *
      * This function is reentrant safe and may be called at any time.
      *
-     * @param mon
      * @see lcb_confmon_add_listener()
      * @see #stop()
      * @see #is_refreshing()
      */
-    void start();
+    void start(bool refresh = false);
 
     /**
      * @brief Cancel a pending configuration refresh.
@@ -227,7 +231,6 @@ struct Confmon {
      * This function is safe to call anywhere. If the monitor is already stopped
      * then this function does nothing.
      *
-     * @param mon
      * @see #start()
      * @see #is_refreshing()
      */
@@ -235,10 +238,10 @@ struct Confmon {
 
     /**
      * @brief Check if the monitor is waiting for a new config from a provider
-     * @param mon
      * @return true if refreshing, false if idle
      */
-    bool is_refreshing() const {
+    bool is_refreshing() const
+    {
         return (state & CONFMON_S_ACTIVE) != 0;
     }
 
@@ -246,7 +249,8 @@ struct Confmon {
      * Get the current configuration
      * @return The configuration
      */
-    ConfigInfo* get_config() const {
+    ConfigInfo *get_config() const
+    {
         return config;
     }
 
@@ -254,23 +258,24 @@ struct Confmon {
      * Get the last error which occurred on this object
      * @return The last error
      */
-    lcb_error_t get_last_error() const {
+    lcb_STATUS get_last_error() const
+    {
         return last_error;
     }
 
     /**
      * @brief Get the current monitor state
-     * @param mon the monitor
      * @return a set of flags consisting of @ref State values.
      */
-    int get_state() const {
+    int get_state() const
+    {
         return state;
     }
 
     void stop_real();
     void do_next_provider();
-    int do_set_next(ConfigInfo*, bool notify_miss);
-    void invoke_listeners(EventType, ConfigInfo*);
+    int do_set_next(ConfigInfo *, bool notify_miss);
+    void invoke_listeners(EventType, ConfigInfo *);
 
     /**
      * @brief Indicate that a provider has failed and advance the monitor
@@ -289,10 +294,10 @@ struct Confmon {
      * provider or enter the inactive state depending on the configuration and
      * whether the current provider is the last provider in the list.
      *
-     * @param provider
-     * @param err
+     * @param which reference to provider, which has been failed
+     * @param why error code
      */
-    void provider_failed(Provider *which, lcb_error_t why);
+    void provider_failed(Provider *which, lcb_STATUS why);
 
     /**
      * @brief Indicate that a provider has successfuly retrieved a configuration.
@@ -300,7 +305,7 @@ struct Confmon {
      * Indicates that the provider has fetched a new configuration from the network
      * and that confmon should attempt to propagate it. It has similar semantics
      * to lcb_confmon_provider_failed() except that the second argument is a config
-     * object rather than an error code. The second argument must not be `NULL`
+     * object rather than an error code. The second argument must not be `nullptr`
      *
      * The monitor will compare the new config against the current config.
      * If the new config does not feature any changes from the current config then
@@ -310,18 +315,19 @@ struct Confmon {
      * This function should _not_ be called outside of an asynchronous provider's
      * handler.
      *
-     * @param provider the provider which yielded the new configuration
-     * @param info the new configuration
+     * @param which the provider which yielded the new configuration
+     * @param config the new configuration
      */
-    void provider_got_config(Provider *which, ConfigInfo* config);
+    void provider_got_config(Provider *which, ConfigInfo *config);
 
-    /** Dump information about the monitor
-     * @param mon the monitor object
+    /**
+     * Dump information about the monitor
      * @param fp the file to which information should be written
      */
     void dump(FILE *fp);
 
-    Provider* get_provider(Method m) const {
+    Provider *get_provider(Method m) const
+    {
         return all_providers[m];
     }
 
@@ -337,16 +343,14 @@ struct Confmon {
      * until remove_listener is called. Note that the listener is not allocated
      * by the confmon and its responsibility is the user's
      *
-     * @param mon the monitor
-     * @param listener the listener. The listener's contents are not copied into
+     * @param lsn the listener. The listener's contents are not copied into
      * confmon and should thus remain valid until it is removed
      */
-    void add_listener(Listener* lsn);
+    void add_listener(Listener *lsn);
 
     /**
      * @brief Unregister (and remove) a listener added via lcb_confmon_add_listener()
-     * @param mon the monitor
-     * @param listener the listener
+     * @param lsn the listener
      */
     void remove_listener(Listener *lsn);
 
@@ -357,18 +361,18 @@ struct Confmon {
     Provider *cur_provider;
 
     /** All providers we know about. Currently this means the 'builtin' providers */
-    Provider* all_providers[CLCONFIG_MAX];
+    Provider *all_providers[CLCONFIG_MAX]{};
 
     /** The current configuration pointer. This contains the most recent accepted
      * configuration */
-    ConfigInfo * config;
+    ConfigInfo *config;
 
-    typedef std::list<Listener*> ListenerList;
+    typedef std::list<Listener *> ListenerList;
     /**  List of listeners for events */
     ListenerList listeners;
 
     lcb_settings *settings;
-    lcb_error_t last_error;
+    lcb_STATUS last_error;
     lcbio_pTABLE iot;
 
     /** This is the async handle for a reentrant start */
@@ -381,10 +385,13 @@ struct Confmon {
     int state;
 
     /** Last time the provider was stopped. As a microsecond timestamp */
-    lcb_uint32_t last_stop_us;
+    lcb_uint64_t last_stop_us;
 
-    typedef std::list<Provider*> ProviderList;
+    typedef std::list<Provider *> ProviderList;
     ProviderList active_providers;
+
+    lcb_INSTANCE *instance;
+    size_t active_provider_list_id;
 };
 
 /**
@@ -392,7 +399,7 @@ struct Confmon {
  * 'subclassed' by implementors.
  */
 struct Provider {
-    Provider(Confmon*, Method type_);
+    Provider(Confmon *, Method type_);
 
     /** Destroy the resources created by this provider. */
     virtual ~Provider();
@@ -403,8 +410,7 @@ struct Provider {
      * this method as an asynchronous return value for a previously-received
      * configuration.
      */
-    virtual ConfigInfo* get_cached() = 0;
-
+    virtual ConfigInfo *get_cached() = 0;
 
     /**
      * Request a new configuration. This will be called by the manager when
@@ -419,10 +425,8 @@ struct Provider {
      * against a provider taking an excessively long time; therefore a provider
      * should implement a timeout mechanism of its choice to promptly deliver
      * a success or failure.
-     *
-     * @param pb
      */
-    virtual lcb_error_t refresh() = 0;
+    virtual lcb_STATUS refresh() = 0;
 
     /**
      * Callback invoked to the provider to indicate that it should cease
@@ -432,55 +436,64 @@ struct Provider {
      * be needed again in quite some time. How long this "time" is can range
      * between 0 seconds and several minutes depending on how a user has
      * configured the client.
-     * @param pb
+     *
      * @return true if actually paused
      */
-    virtual bool pause() {
+    virtual bool pause()
+    {
         return false;
     }
 
     /**
      * Called when a new configuration has been received.
      *
-     * @param provider the provider instance
      * @param config the current configuration.
      * Note that this should only update the server list and do nothing
      * else.
      */
-    virtual void config_updated(lcbvb_CONFIG*) {
+    virtual void config_updated(lcbvb_CONFIG *config)
+    {
+        (void)config;
     }
 
     /**
      * Retrieve the list of nodes from this provider, if applicable
-     * @param p the provider
-     * @return A list of nodes, or NULL if the provider does not have a list
+     *
+     * @return A list of nodes, or nullptr if the provider does not have a list
      */
-    virtual const lcb::Hostlist* get_nodes() const {
-        return NULL;
+    virtual const lcb::Hostlist *get_nodes() const
+    {
+        return nullptr;
     }
 
     /**
      * Call to change the configured nodes of this provider.
-     * @param p The provider
+     *
      * @param l The list of nodes to apply
      */
-    virtual void configure_nodes(const lcb::Hostlist&) {
+    virtual void configure_nodes(const lcb::Hostlist &l)
+    {
+        (void)l;
     }
 
     /**
      * Dump state information. This callback is optional
-     * @param p the provider
+     *
      * @param f the file to write to
      */
-    virtual void dump(FILE *) const {
+    virtual void dump(FILE *f) const
+    {
+        (void)f;
     }
 
-    void enable() {
-        enabled = 1;
+    void enable()
+    {
+        enabled = true;
     }
 
-    virtual void enable(void *) {
-        assert("Must be implemented in subclass if used" && 0);
+    virtual void enable(void *)
+    {
+        lcb_assert("Must be implemented in subclass if used" && 0);
     }
 
     /** The type of provider */
@@ -492,31 +505,34 @@ struct Provider {
     /** The parent manager object */
     Confmon *parent;
 
-    lcb_settings& settings() const {
+    lcb_settings &settings() const
+    {
         return *parent->settings;
     }
 };
 
-Provider *new_cccp_provider(Confmon*);
-Provider *new_file_provider(Confmon*);
-Provider *new_http_provider(Confmon*);
-Provider *new_mcraw_provider(Confmon*);
-
+Provider *new_cccp_provider(Confmon *);
+Provider *new_file_provider(Confmon *);
+Provider *new_http_provider(Confmon *);
+Provider *new_mcraw_provider(Confmon *);
+Provider *new_cladmin_provider(Confmon *);
 
 /** @brief refcounted object encapsulating a vbucket config */
-class ConfigInfo {
-public:
+class ConfigInfo
+{
+  public:
     /**
      * Creates a new configuration wrapper object containing the vbucket config
      * pointed to by 'config'. Its initial refcount will be set to 1.
      *
-     * @param config a newly parsed configuration
+     * @param vbc a newly parsed configuration
      * @param origin the type of provider from which the config originated.
      * @return a new ConfigInfo object. This should be incref()'d/decref()'d
      * as needed.
      */
-    static ConfigInfo* create(lcbvb_CONFIG *vbc, Method origin) {
-        return new ConfigInfo(vbc, origin);
+    static ConfigInfo *create(lcbvb_CONFIG *vbc, Method origin, std::string address)
+    {
+        return new ConfigInfo(vbc, origin, std::move(address));
     }
     /**
      * @brief Compares two info structures and determine which one is newer
@@ -524,43 +540,53 @@ public:
      * This function returns an integer less than
      * zero, zero or greater than zero if the first argument is considered older
      * than, equal to, or later than the second argument.
-     * @param a
-     * @param b
+     *
+     * @param config anoother config
      * @see lcbvb_get_revision
      * @see ConfigInfo::cmpclock
      */
-    int compare(const ConfigInfo&);
+    int compare(const ConfigInfo &config, lcbvb_CHANGETYPE chstatus) const;
 
     /**
      * @brief Increment the refcount on a config object
      */
-    void incref() { refcount++; }
+    void incref()
+    {
+        refcount++;
+    }
 
     /**
      * @brief Decrement the refcount on a config object.
      * Decrement the refcount. If the internal refcount reaches 0 then the internal
      * members (including the vbucket config handle itself) will be freed.
-     * @param info the configuration
      */
-    void decref() {
+    void decref()
+    {
         if (!--refcount) {
             delete this;
         }
     }
 
-    operator lcbvb_CONFIG*() const {
+    explicit operator lcbvb_CONFIG *() const
+    {
         return vbc;
     }
 
-    Method get_origin() const {
+    Method get_origin() const
+    {
         return origin;
     }
 
-    /** Actual configuration */
-    lcbvb_CONFIG* vbc;
+    const std::string &get_address() const
+    {
+        return address;
+    }
 
-private:
-    ConfigInfo(lcbvb_CONFIG *vbc, Method origin);
+    /** Actual configuration */
+    lcbvb_CONFIG *vbc;
+
+  private:
+    ConfigInfo(lcbvb_CONFIG *vbc, Method origin, std::string address);
 
     ~ConfigInfo();
 
@@ -572,6 +598,8 @@ private:
 
     /** Origin provider type which produced this config */
     Method origin;
+    /** The address of the node, that sent this config */
+    std::string address{};
 };
 
 /**
@@ -581,18 +609,17 @@ private:
  * for a variety of events the listener can know.
  */
 struct Listener {
-    virtual ~Listener() {
-    }
+    virtual ~Listener() = default;
 
     /** Linked list node */
-    lcb_list_t ll;
+    lcb_list_t ll{};
 
     /**
      * Callback invoked for significant events
      *
      * @param event the event which took place
      * @param config the configuration associated with the event. Note that
-     * `config` may also be NULL
+     * `config` may also be nullptr
      */
     virtual void clconfig_lsn(EventType event, ConfigInfo *config) = 0;
 };
@@ -608,7 +635,7 @@ struct Listener {
  * Sets the input/output filename for the file provider. This also enables
  * the file provider.
  * @param p the provider
- * @param f the filename (if NULL, a temporary filename will be created)
+ * @param f the filename (if nullptr, a temporary filename will be created)
  * @param ro whether the client will never modify the file
  * @return true on success, false on failure.
  */
@@ -616,10 +643,10 @@ bool file_set_filename(Provider *p, const char *f, bool ro);
 
 /**
  * Retrieve the filename for the provider
- * @param p The provider of type LCB_CLCONFIG_FILE
+ * @param p The provider of type CLCONFIG_FILE
  * @return the current filename being used.
  */
-const char* file_get_filename(Provider *p);
+const char *file_get_filename(Provider *p);
 void file_set_readonly(Provider *p, bool val);
 /**@}*/
 
@@ -631,22 +658,24 @@ void file_set_readonly(Provider *p, bool val);
 /**
  * Get the socket representing the current REST connection to the cluster
  * (if applicable)
- * @param mon
+ * @param p The provider of type CLCONFIG_HTTP
  * @return
  */
-const lcbio_SOCKET* http_get_conn(const Provider *p);
+const lcbio_SOCKET *http_get_conn(const Provider *p);
 
-static inline const lcbio_SOCKET* http_get_conn(Confmon *c) {
+static inline const lcbio_SOCKET *http_get_conn(Confmon *c)
+{
     return http_get_conn(c->get_provider(CLCONFIG_HTTP));
 }
 
 /**
  * Get the hostname for the current REST connection to the cluster
- * @param mon
+ * @param p The provider of type CLCONFIG_HTTP
  * @return
  */
-const lcb_host_t * http_get_host(const Provider *p);
-static inline const lcb_host_t* http_get_host(Confmon *c) {
+const lcb_host_t *http_get_host(const Provider *p);
+static inline const lcb_host_t *http_get_host(Confmon *c)
+{
     return http_get_host(c->get_provider(CLCONFIG_HTTP));
 }
 /**@}*/
@@ -674,8 +703,7 @@ static inline const lcb_host_t* http_get_host(Confmon *c) {
  * @return LCB_SUCCESS, or an error code if the configuration could not be
  * set
  */
-lcb_error_t
-cccp_update(Provider *provider, const char *host, const char *data);
+lcb_STATUS cccp_update(Provider *provider, const char *host, const char *data);
 
 /**
  * @brief Notify the CCCP provider about a configuration received from a
@@ -687,8 +715,14 @@ cccp_update(Provider *provider, const char *host, const char *data);
  * @param nbytes Size of payload
  * @param origin Host object from which the packet was received
  */
-void cccp_update(const void *cookie, lcb_error_t err,
-    const void *bytes, size_t nbytes, const lcb_host_t *origin);
+void cccp_update(const void *cookie, lcb_STATUS err, const void *bytes, size_t nbytes, const lcb_host_t *origin);
+
+/**
+ * @brief record status of SELECT_BUCKET command
+ * @param cookie_
+ * @param err
+ */
+void select_status(const void *cookie_, lcb_STATUS err);
 
 /**@}*/
 
@@ -697,6 +731,6 @@ void cccp_update(const void *cookie, lcb_error_t err,
 /**@}*/
 /**@}*/
 
-} // clconfig
-} // lcb
+} // namespace clconfig
+} // namespace lcb
 #endif /* LCB_CLCONFIG_H */
