@@ -1,6 +1,6 @@
 /* -*- Mode: C++; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 /*
- *     Copyright 2016 Couchbase, Inc.
+ *     Copyright 2016-2020 Couchbase, Inc.
  *
  *   Licensed under the Apache License, Version 2.0 (the "License");
  *   you may not use this file except in compliance with the License.
@@ -26,22 +26,23 @@
 #include "settings.h"
 #include "internal.h"
 
+#include "capi/query.hh"
+
 #define LOGFMT "(mgreq=%p) "
-#define LOGID(req) static_cast<const void*>(req)
+#define LOGID(req) static_cast<const void *>(req)
 #define LOGARGS(req, lvl) (req)->m_instance->settings, "ixmgmt", LCB_LOG_##lvl, __FILE__, __LINE__
 
-using std::vector;
 using std::string;
+using std::vector;
 
-static const char *
-ixtype_2_str(unsigned ixtype)
+static const char *ixtype_2_str(unsigned ixtype)
 {
     if (ixtype == LCB_N1XSPEC_T_GSI) {
         return "gsi";
     } else if (ixtype == LCB_N1XSPEC_T_VIEW) {
         return "view";
     } else {
-        return NULL;
+        return nullptr;
     }
 }
 
@@ -52,37 +53,38 @@ struct IndexOpCtx {
 
 struct ErrorSpec {
     string msg;
-    unsigned code;
+    unsigned code{};
 };
 
-template <typename T> void my_delete(T p) {
+template <typename T>
+void my_delete(T p)
+{
     delete p;
 }
 
-template <typename T> lcb_error_t
-extract_n1ql_errors(const char *s, size_t n, T& err_out)
+template <typename T>
+lcb_STATUS extract_n1ql_errors(const char *s, size_t n, T &err_out)
 {
     Json::Value jresp;
     if (!Json::Reader().parse(s, s + n, jresp)) {
-        return LCB_PROTOCOL_ERROR;
+        return LCB_ERR_PROTOCOL_ERROR;
     }
     if (jresp["status"].asString() == "success") {
         return LCB_SUCCESS;
     }
 
-    Json::Value& errors = jresp["errors"];
+    Json::Value &errors = jresp["errors"];
     if (errors.isNull()) {
         return LCB_SUCCESS;
     } else if (!errors.isArray()) {
-        return LCB_PROTOCOL_ERROR;
+        return LCB_ERR_PROTOCOL_ERROR;
     }
 
     if (errors.empty()) {
         return LCB_SUCCESS;
     }
 
-    for (Json::ArrayIndex ii = 0; ii < errors.size(); ++ii) {
-        const Json::Value& err = errors[ii];
+    for (auto &err : errors) {
         if (!err.isObject()) {
             continue; // expected an object!
         }
@@ -91,44 +93,42 @@ extract_n1ql_errors(const char *s, size_t n, T& err_out)
         spec.code = err["code"].asUInt();
         err_out.insert(err_out.end(), spec);
     }
-    return LCB_ERROR;
+    return LCB_ERR_GENERIC;
 }
 
-static lcb_error_t
-get_n1ql_error(const char *s, size_t n)
+static lcb_STATUS get_n1ql_error(const char *s, size_t n)
 {
     std::vector<ErrorSpec> dummy;
     return extract_n1ql_errors(s, n, dummy);
 }
 
 // Called for generic operations and establishes existence or lack thereof
-static void
-cb_generic(lcb_t instance, int, const lcb_RESPN1QL *resp)
+static void cb_generic(lcb_INSTANCE *instance, int, const lcb_RESPQUERY *resp)
 {
     // Get the real cookie..
     if (!(resp->rflags & LCB_RESP_F_FINAL)) {
         return;
     }
 
-    IndexOpCtx *ctx = reinterpret_cast<IndexOpCtx*>(resp->cookie);
-    lcb_RESPN1XMGMT w_resp = { 0 };
+    auto *ctx = reinterpret_cast<IndexOpCtx *>(resp->cookie);
+    lcb_RESPN1XMGMT w_resp{};
     w_resp.cookie = ctx->cookie;
 
-    if ((w_resp.rc = resp->rc) == LCB_SUCCESS || resp->rc == LCB_HTTP_ERROR) {
+    if ((w_resp.rc = resp->ctx.rc) == LCB_SUCCESS || resp->ctx.rc == LCB_ERR_HTTP) {
         // Check if the top-level N1QL response succeeded, and then
         // descend to determine additional errors. This is primarily
         // required to support EEXIST for GSI primary indexes
 
         vector<ErrorSpec> errors;
-        lcb_error_t rc = extract_n1ql_errors(resp->row, resp->nrow, errors);
-        if (rc == LCB_ERROR) {
-            w_resp.rc = LCB_QUERY_ERROR;
-            for (size_t ii = 0; ii < errors.size(); ++ii) {
-                const std::string& msg = errors[ii].msg;
+        lcb_STATUS rc = extract_n1ql_errors(resp->row, resp->nrow, errors);
+        if (rc == LCB_ERR_GENERIC) {
+            w_resp.rc = LCB_ERR_QUERY;
+            for (auto &error : errors) {
+                const std::string &msg = error.msg;
                 if (msg.find("already exist") != string::npos) {
-                    w_resp.rc = LCB_KEY_EEXISTS; // Index entry already exists
+                    w_resp.rc = LCB_ERR_DOCUMENT_EXISTS; // Index entry already exists
                 } else if (msg.find("not found") != string::npos) {
-                    w_resp.rc = LCB_KEY_ENOENT;
+                    w_resp.rc = LCB_ERR_DOCUMENT_NOT_FOUND;
                 }
             }
         } else {
@@ -137,7 +137,7 @@ cb_generic(lcb_t instance, int, const lcb_RESPN1QL *resp)
     }
 
     w_resp.inner = resp;
-    w_resp.specs = NULL;
+    w_resp.specs = nullptr;
     w_resp.nspecs = 0;
     ctx->callback(instance, LCB_CALLBACK_IXMGMT, &w_resp);
     delete ctx;
@@ -152,72 +152,81 @@ cb_generic(lcb_t instance, int, const lcb_RESPN1QL *resp)
  *        is done
  * @param s N1QL request payload
  * @param n N1QL request length
- * @param obj Internal context. Created with new if NULL
+ * @param obj Internal context. Created with new if nullptr
  * @return
  *
  * See other overload for passing just the query string w/o extra parameters
  */
-template <typename T> lcb_error_t
-dispatch_common(lcb_t instance,
-    const void *cookie, lcb_N1XMGMTCALLBACK u_callback,
-    lcb_N1QLCALLBACK i_callback, const char *s, size_t n, T *obj)
+template <typename T>
+lcb_STATUS dispatch_common(lcb_INSTANCE *instance, const void *cookie, lcb_N1XMGMTCALLBACK u_callback,
+                           lcb_QUERY_CALLBACK i_callback, const char *s, size_t n, T *obj)
 {
-    lcb_error_t rc = LCB_SUCCESS;
     bool our_alloc = false;
-    lcb_CMDN1QL cmd = { 0 };
-    struct { lcb_t m_instance; } ixwrap = { instance }; // For logging
+    struct {
+        lcb_INSTANCE *m_instance;
+    } ixwrap = {instance}; // For logging
 
-    if (obj == NULL) {
+    if (obj == nullptr) {
         obj = new T();
         our_alloc = true;
     }
 
+    lcb_STATUS rc;
     if (!(obj->callback = u_callback)) {
-        rc = LCB_EINVAL;
+        rc = LCB_ERR_INVALID_ARGUMENT;
         goto GT_ERROR;
     }
 
-    obj->cookie = const_cast<void*>(cookie);
+    obj->cookie = const_cast<void *>(cookie);
 
-    cmd.query = s;
-    cmd.nquery = n;
-    cmd.callback = i_callback;
+    lcb_CMDQUERY *cmd;
+    lcb_cmdquery_create(&cmd);
+    lcb_cmdquery_payload(cmd, s, n);
+    lcb_cmdquery_callback(cmd, i_callback);
     lcb_log(LOGARGS(&ixwrap, DEBUG), LOGFMT "Issuing query %.*s", LOGID(obj), (int)n, s);
-    rc = lcb_n1ql_query(instance, obj, &cmd);
+    rc = lcb_query(instance, obj, cmd);
+    lcb_cmdquery_destroy(cmd);
 
-    GT_ERROR:
+GT_ERROR:
     if (rc != LCB_SUCCESS && our_alloc) {
         delete obj;
     }
     return rc;
 }
 
-template <typename T> lcb_error_t
-dispatch_common(lcb_t instance,
-    const void *cookie, lcb_N1XMGMTCALLBACK u_callback,
-    lcb_N1QLCALLBACK i_callback, const string& ss, T *obj = NULL)
+template <typename T>
+lcb_STATUS dispatch_common(lcb_INSTANCE *instance, const void *cookie, lcb_N1XMGMTCALLBACK u_callback,
+                           lcb_QUERY_CALLBACK i_callback, const string &ss, T *obj = nullptr)
 {
     Json::Value root;
     root["statement"] = ss;
     string reqbuf = Json::FastWriter().write(root);
-    return dispatch_common<T>(instance,
-        cookie, u_callback, i_callback,
-        reqbuf.c_str(), reqbuf.size()-1 /*newline*/, obj);
+    return dispatch_common<T>(instance, cookie, u_callback, i_callback, reqbuf.c_str(), reqbuf.size() - 1 /*newline*/,
+                              obj);
 }
 
-
 // Class to back the storage for the actual lcb_IXSPEC without doing too much
 // mind-numbing buffer copies. Maybe this can be done via a macro instead?
-class IndexSpec : public lcb_N1XSPEC {
-public:
-    IndexSpec(const char *s, size_t n) : lcb_N1XSPEC() {
+class IndexSpec : public lcb_N1XSPEC
+{
+  public:
+    IndexSpec(const char *s, size_t n) : lcb_N1XSPEC()
+    {
         load_json(s, n);
     }
-    inline IndexSpec(const lcb_N1XSPEC *spec);
-    static inline void to_key(const lcb_N1XSPEC *spec, std::string& out);
-    bool is_primary() const { return flags & LCB_N1XSPEC_F_PRIMARY; }
-    bool is_defer() const { return flags & LCB_N1XSPEC_F_DEFER; }
-    void ensure_keyspace(lcb_t instance) {
+    IndexSpec(const IndexSpec &) = delete;
+    inline explicit IndexSpec(const lcb_N1XSPEC *spec);
+    static inline void to_key(const lcb_N1XSPEC *spec, std::string &out);
+    bool is_primary() const
+    {
+        return flags & LCB_N1XSPEC_F_PRIMARY;
+    }
+    bool is_defer() const
+    {
+        return flags & LCB_N1XSPEC_F_DEFER;
+    }
+    void ensure_keyspace(lcb_INSTANCE *instance)
+    {
         if (nkeyspace) {
             return;
         }
@@ -225,39 +234,38 @@ public:
         nkeyspace = strlen(keyspace);
     }
 
-private:
+  private:
     // Load fields from a JSON string
     inline void load_json(const char *s, size_t n);
 
     // Load all fields
-    inline size_t load_fields(const Json::Value& root, bool do_copy);
+    inline size_t load_fields(const Json::Value &root, bool do_copy);
 
-    size_t total_fields_size(const Json::Value& src) {
+    size_t total_fields_size(const Json::Value &src)
+    {
         return load_fields(src, false);
     }
 
     // Load field from a JSON object
-    inline size_t load_json_field(
-        const Json::Value& root,
-        const char *name, const char **tgt_ptr, size_t *tgt_len, bool do_copy);
+    inline size_t load_json_field(const Json::Value &root, const char *name, const char **tgt_ptr, size_t *tgt_len,
+                                  bool do_copy);
 
     // Load field from another pointer
-    void load_field(const char **dest, const char *src, size_t n) {
+    void load_field(const char **dest, const char *src, size_t n)
+    {
         m_buf.append(src, n);
         if (n) {
-            *dest = &m_buf.c_str()[m_buf.size()-n];
+            *dest = &m_buf.c_str()[m_buf.size() - n];
         } else {
-            *dest = NULL;
+            *dest = nullptr;
         }
     }
 
     string m_buf;
-    IndexSpec(const IndexSpec&);
 };
 
 LIBCOUCHBASE_API
-lcb_error_t
-lcb_n1x_create(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
+lcb_STATUS lcb_n1x_create(lcb_INSTANCE *instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
 {
     string ss;
     IndexSpec spec(&cmd->spec);
@@ -267,7 +275,7 @@ lcb_n1x_create(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
     if (spec.is_primary()) {
         ss += " PRIMARY";
     } else if (!spec.nname) {
-        return LCB_EMPTY_KEY;
+        return LCB_ERR_EMPTY_KEY;
     }
     ss.append(" INDEX");
     if (spec.nname) {
@@ -277,7 +285,7 @@ lcb_n1x_create(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
 
     if (!spec.is_primary()) {
         if (!spec.nfields) {
-            return LCB_EMPTY_KEY;
+            return LCB_ERR_EMPTY_KEY;
         }
 
         // See if we can parse 'fields' properly. First, try to parse as
@@ -286,40 +294,40 @@ lcb_n1x_create(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
         Json::Reader r;
         if (!r.parse(spec.fields, spec.fields + spec.nfields, fields_arr)) {
             // Invalid JSON!
-            return LCB_EINVAL;
+            return LCB_ERR_INVALID_ARGUMENT;
         }
 
         ss.append(" (");
         if (fields_arr.isArray()) {
-            if (!fields_arr.size()) {
-                return LCB_EMPTY_KEY;
+            if (fields_arr.empty()) {
+                return LCB_ERR_EMPTY_KEY;
             }
             for (size_t ii = 0; ii < fields_arr.size(); ++ii) {
                 static Json::Value empty;
-                const Json::Value& field = fields_arr.get(ii, empty);
+                const Json::Value &field = fields_arr.get(ii, empty);
                 if (!field.isString()) {
-                    return LCB_EINVAL;
+                    return LCB_ERR_INVALID_ARGUMENT;
                 }
                 ss.append(field.asString());
-                if (ii != fields_arr.size()-1) {
+                if (ii != fields_arr.size() - 1) {
                     ss.append(",");
                 }
             }
         } else if (fields_arr.isString()) {
             std::string field_list = fields_arr.asString();
             if (field_list.empty()) {
-                return LCB_EMPTY_KEY;
+                return LCB_ERR_EMPTY_KEY;
             }
             ss.append(field_list);
         } else {
-            return LCB_EINVAL;
+            return LCB_ERR_INVALID_ARGUMENT;
         }
         ss.append(") ");
     }
 
     if (spec.ncond) {
         if (spec.is_primary()) {
-            return LCB_EINVAL;
+            return LCB_ERR_INVALID_ARGUMENT;
         }
         ss.append(" WHERE ").append(spec.cond, spec.ncond).append(" ");
     }
@@ -327,7 +335,7 @@ lcb_n1x_create(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
     if (spec.ixtype) {
         const char *ixtype = ixtype_2_str(spec.ixtype);
         if (!ixtype) {
-            return LCB_EINVAL;
+            return LCB_ERR_INVALID_ARGUMENT;
         }
         ss.append(" USING ").append(ixtype);
     }
@@ -339,57 +347,58 @@ lcb_n1x_create(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
     return dispatch_common<IndexOpCtx>(instance, cookie, cmd->callback, cb_generic, ss);
 }
 
+class ListIndexCtx : public IndexOpCtx
+{
+  public:
+    vector<IndexSpec *> specs;
 
-class ListIndexCtx : public IndexOpCtx {
-public:
-    vector<IndexSpec*> specs;
-
-    virtual void invoke(lcb_t instance, lcb_RESPN1XMGMT *resp) {
+    virtual void invoke(lcb_INSTANCE *instance, lcb_RESPN1XMGMT *resp)
+    {
         finish(instance, resp);
     }
 
-    void finish(lcb_t instance, lcb_RESPN1XMGMT *resp = NULL) {
-        lcb_RESPN1XMGMT w_resp = { 0 };
-        if (resp == NULL) {
+    void finish(lcb_INSTANCE *instance, lcb_RESPN1XMGMT *resp = nullptr)
+    {
+        lcb_RESPN1XMGMT w_resp{};
+        if (resp == nullptr) {
             resp = &w_resp;
             resp->rc = LCB_SUCCESS;
         }
         resp->cookie = cookie;
-        lcb_N1XSPEC **speclist = reinterpret_cast<lcb_N1XSPEC**>(&specs[0]);
+        auto **speclist = reinterpret_cast<lcb_N1XSPEC **>(&specs[0]);
         resp->specs = speclist;
         resp->nspecs = specs.size();
         callback(instance, LCB_CALLBACK_IXMGMT, resp);
         delete this;
     }
 
-    virtual ~ListIndexCtx() {
-        for (size_t ii = 0; ii < specs.size(); ++ii) {
-            delete specs[ii];
+    virtual ~ListIndexCtx()
+    {
+        for (auto &spec : specs) {
+            delete spec;
         }
         specs.clear();
     }
 };
 
-static void
-cb_index_list(lcb_t instance, int, const lcb_RESPN1QL *resp)
+static void cb_index_list(lcb_INSTANCE *instance, int, const lcb_RESPQUERY *resp)
 {
-    ListIndexCtx *ctx = reinterpret_cast<ListIndexCtx *>(resp->cookie);
+    auto *ctx = reinterpret_cast<ListIndexCtx *>(resp->cookie);
     if (!(resp->rflags & LCB_RESP_F_FINAL)) {
         ctx->specs.push_back(new IndexSpec(resp->row, resp->nrow));
         return;
     }
 
-    lcb_RESPN1XMGMT w_resp = { 0 };
-    if ((w_resp.rc = resp->rc) == LCB_SUCCESS) {
+    lcb_RESPN1XMGMT w_resp{};
+    if ((w_resp.rc = resp->ctx.rc) == LCB_SUCCESS) {
         w_resp.rc = get_n1ql_error(resp->row, resp->nrow);
     }
     w_resp.inner = resp;
     ctx->invoke(instance, &w_resp);
 }
 
-static lcb_error_t
-do_index_list(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd,
-    ListIndexCtx *ctx)
+static lcb_STATUS do_index_list(lcb_INSTANCE *instance, const void *cookie, const lcb_CMDN1XMGMT *cmd,
+                                ListIndexCtx *ctx)
 {
     string ss;
     IndexSpec spec(&cmd->spec);
@@ -406,11 +415,8 @@ do_index_list(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd,
     }
     if (spec.ixtype) {
         const char *s_ixtype = ixtype_2_str(spec.ixtype);
-        if (s_ixtype == NULL) {
-            if (ctx != NULL) {
-                delete ctx;
-            }
-            return LCB_EINVAL;
+        if (s_ixtype == nullptr) {
+            return LCB_ERR_INVALID_ARGUMENT;
         }
         ss.append(" using=\"").append(s_ixtype).append("\" AND");
     }
@@ -422,20 +428,17 @@ do_index_list(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd,
     ss.append(" true");
     ss.append(" ORDER BY is_primary DESC, name ASC");
 
-    return dispatch_common<ListIndexCtx>(instance,
-        cookie, cmd->callback, cb_index_list, ss, ctx);
+    return dispatch_common<ListIndexCtx>(instance, cookie, cmd->callback, cb_index_list, ss, ctx);
 }
 
 LIBCOUCHBASE_API
-lcb_error_t
-lcb_n1x_list(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
+lcb_STATUS lcb_n1x_list(lcb_INSTANCE *instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
 {
-    return do_index_list(instance, cookie, cmd, NULL);
+    return do_index_list(instance, cookie, cmd, nullptr);
 }
 
 LIBCOUCHBASE_API
-lcb_error_t
-lcb_n1x_drop(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
+lcb_STATUS lcb_n1x_drop(lcb_INSTANCE *instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
 {
     string ss;
     IndexSpec spec(&cmd->spec);
@@ -449,13 +452,13 @@ lcb_n1x_drop(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
         ss = "DROP PRIMARY INDEX ON";
         ss.append(" `").append(spec.keyspace, spec.nkeyspace).append("`");
     } else {
-        return LCB_EMPTY_KEY;
+        return LCB_ERR_EMPTY_KEY;
     }
 
     if (spec.ixtype) {
         const char *stype = ixtype_2_str(spec.ixtype);
         if (!stype) {
-            return LCB_EINVAL;
+            return LCB_ERR_INVALID_ARGUMENT;
         }
         ss.append(" USING ").append(stype);
     }
@@ -463,40 +466,37 @@ lcb_n1x_drop(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
     return dispatch_common<IndexOpCtx>(instance, cookie, cmd->callback, cb_generic, ss);
 }
 
-class ListIndexCtx_BuildIndex : public ListIndexCtx {
-public:
-    virtual inline void invoke(lcb_t instance, lcb_RESPN1XMGMT *resp);
-    inline lcb_error_t try_build(lcb_t instance);
+class ListIndexCtx_BuildIndex : public ListIndexCtx
+{
+  public:
+    inline void invoke(lcb_INSTANCE *instance, lcb_RESPN1XMGMT *resp) override;
+    inline lcb_STATUS try_build(lcb_INSTANCE *instance);
 };
 
-static void
-cb_build_submitted(lcb_t instance, int, const lcb_RESPN1QL *resp)
+static void cb_build_submitted(lcb_INSTANCE *instance, int, const lcb_RESPQUERY *resp)
 {
-    ListIndexCtx *ctx = reinterpret_cast<ListIndexCtx*>(resp->cookie);
+    auto *ctx = reinterpret_cast<ListIndexCtx *>(resp->cookie);
 
     if (resp->rflags & LCB_RESP_F_FINAL) {
-        lcb_RESPN1XMGMT w_resp = { 0 };
-        if ((w_resp.rc = resp->rc) == LCB_SUCCESS) {
+        lcb_RESPN1XMGMT w_resp{};
+        if ((w_resp.rc = resp->ctx.rc) == LCB_SUCCESS) {
             w_resp.rc = get_n1ql_error(resp->row, resp->nrow);
         }
         ctx->finish(instance, &w_resp);
     }
 }
 
-lcb_error_t
-ListIndexCtx_BuildIndex::try_build(lcb_t instance)
+lcb_STATUS ListIndexCtx_BuildIndex::try_build(lcb_INSTANCE *instance)
 {
-    vector<IndexSpec*> pending;
-    for (size_t ii = 0; ii < specs.size(); ++ii) {
-        IndexSpec* spec = specs[ii];
-        if (strncmp(spec->state, "pending", spec->nstate) == 0 ||
-                strncmp(spec->state, "deferred", spec->nstate) == 0) {
+    vector<IndexSpec *> pending;
+    for (auto spec : specs) {
+        if (strncmp(spec->state, "pending", spec->nstate) == 0 || strncmp(spec->state, "deferred", spec->nstate) == 0) {
             pending.push_back(spec);
         }
     }
 
     if (pending.empty()) {
-        return LCB_KEY_ENOENT;
+        return LCB_ERR_DOCUMENT_NOT_FOUND;
     }
 
     string ss;
@@ -508,45 +508,40 @@ ListIndexCtx_BuildIndex::try_build(lcb_t instance)
         ss += '`';
         ss.append(pending[ii]->name, pending[ii]->nname);
         ss += '`';
-        if (ii+1 < pending.size()) {
+        if (ii + 1 < pending.size()) {
             ss += ',';
         }
     }
     ss += ')';
 
-    lcb_error_t rc = dispatch_common<ListIndexCtx_BuildIndex>(
-        instance, cookie, callback, cb_build_submitted, ss,
-        this);
+    lcb_STATUS rc = dispatch_common<ListIndexCtx_BuildIndex>(instance, cookie, callback, cb_build_submitted, ss, this);
 
     if (rc == LCB_SUCCESS) {
-        std::set<IndexSpec*> to_remove(specs.begin(), specs.end());
-        for (size_t ii = 0; ii < pending.size(); ++ii) {
-            to_remove.erase(pending[ii]);
+        std::set<IndexSpec *> to_remove(specs.begin(), specs.end());
+        for (auto &ii : pending) {
+            to_remove.erase(ii);
         }
 
-        std::for_each(to_remove.begin(), to_remove.end(), my_delete<IndexSpec*>);
+        std::for_each(to_remove.begin(), to_remove.end(), my_delete<IndexSpec *>);
 
         specs = pending;
     }
     return rc;
 }
 
-void
-ListIndexCtx_BuildIndex::invoke(lcb_t instance, lcb_RESPN1XMGMT *resp)
+void ListIndexCtx_BuildIndex::invoke(lcb_INSTANCE *instance, lcb_RESPN1XMGMT *resp)
 {
-    if (resp->rc == LCB_SUCCESS &&
-            (resp->rc = try_build(instance)) == LCB_SUCCESS) {
+    if (resp->rc == LCB_SUCCESS && (resp->rc = try_build(instance)) == LCB_SUCCESS) {
         return;
     }
     finish(instance, resp);
 }
 
 LIBCOUCHBASE_API
-lcb_error_t
-lcb_n1x_startbuild(lcb_t instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
+lcb_STATUS lcb_n1x_startbuild(lcb_INSTANCE *instance, const void *cookie, const lcb_CMDN1XMGMT *cmd)
 {
-    ListIndexCtx_BuildIndex *ctx = new ListIndexCtx_BuildIndex();
-    lcb_error_t rc = do_index_list(instance, cookie, cmd, ctx);
+    auto *ctx = new ListIndexCtx_BuildIndex();
+    lcb_STATUS rc = do_index_list(instance, cookie, cmd, ctx);
     if (rc != LCB_SUCCESS) {
         delete ctx;
     }
@@ -558,26 +553,25 @@ struct WatchIndexCtx : public IndexOpCtx {
     lcbio_pTIMER m_timer;
     uint32_t m_interval;
     uint64_t m_tsend;
-    lcb_t m_instance;
-    std::map<std::string,IndexSpec*> m_defspend;
-    std::vector<IndexSpec*> m_defsok;
+    lcb_INSTANCE *m_instance;
+    std::map<std::string, IndexSpec *> m_defspend;
+    std::vector<IndexSpec *> m_defsok;
 
     inline void read_state(const lcb_RESPN1XMGMT *resp);
     inline void reschedule();
-    inline lcb_error_t do_poll();
-    inline lcb_error_t load_defs(const lcb_CMDN1XWATCH *);
-    inline WatchIndexCtx(lcb_t, const void *, const lcb_CMDN1XWATCH*);
+    inline lcb_STATUS do_poll();
+    inline lcb_STATUS load_defs(const lcb_CMDN1XWATCH *);
+    inline WatchIndexCtx(lcb_INSTANCE *, const void *, const lcb_CMDN1XWATCH *);
     inline ~WatchIndexCtx();
-    inline void finish(lcb_error_t rc, const lcb_RESPN1XMGMT *);
+    inline void finish(lcb_STATUS rc, const lcb_RESPN1XMGMT *);
 };
 
-static void
-cb_watchix_tm(void *arg)
+static void cb_watchix_tm(void *arg)
 {
-    WatchIndexCtx *ctx = reinterpret_cast<WatchIndexCtx*>(arg);
+    auto *ctx = reinterpret_cast<WatchIndexCtx *>(arg);
     uint64_t now = lcb_nstime();
     if (now >= ctx->m_tsend) {
-        ctx->finish(LCB_ETIMEDOUT, NULL);
+        ctx->finish(LCB_ERR_TIMEOUT, nullptr);
     } else {
         ctx->do_poll();
     }
@@ -586,8 +580,8 @@ cb_watchix_tm(void *arg)
 #define DEFAULT_WATCH_TIMEOUT LCB_S2US(30)
 #define DEFAULT_WATCH_INTERVAL LCB_MS2US(500)
 
-WatchIndexCtx::WatchIndexCtx(lcb_t instance, const void *cookie_, const lcb_CMDN1XWATCH *cmd)
-: m_instance(instance)
+WatchIndexCtx::WatchIndexCtx(lcb_INSTANCE *instance, const void *cookie_, const lcb_CMDN1XWATCH *cmd)
+    : IndexOpCtx(), m_instance(instance)
 {
     uint64_t now = lcb_nstime();
     uint32_t timeout = cmd->timeout ? cmd->timeout : DEFAULT_WATCH_TIMEOUT;
@@ -596,10 +590,10 @@ WatchIndexCtx::WatchIndexCtx(lcb_t instance, const void *cookie_, const lcb_CMDN
     m_tsend = now + LCB_US2NS(timeout);
 
     this->callback = cmd->callback;
-    this->cookie = const_cast<void*>(cookie_);
+    this->cookie = const_cast<void *>(cookie_);
 
     m_timer = lcbio_timer_new(instance->iotable, this, cb_watchix_tm);
-    lcb_aspend_add(&instance->pendops, LCB_PENDTYPE_COUNTER, NULL);
+    lcb_aspend_add(&instance->pendops, LCB_PENDTYPE_COUNTER, nullptr);
 }
 
 WatchIndexCtx::~WatchIndexCtx()
@@ -608,34 +602,31 @@ WatchIndexCtx::~WatchIndexCtx()
         lcbio_timer_destroy(m_timer);
     }
     if (m_instance) {
-        lcb_aspend_del(&m_instance->pendops, LCB_PENDTYPE_COUNTER, NULL);
+        lcb_aspend_del(&m_instance->pendops, LCB_PENDTYPE_COUNTER, nullptr);
         lcb_maybe_breakout(m_instance);
     }
 
-    std::for_each(m_defsok.begin(), m_defsok.end(), my_delete<IndexSpec*>);
-    for (std::map<string,IndexSpec*>::iterator ii = m_defspend.begin();
-            ii != m_defspend.end(); ++ii) {
-        delete ii->second;
+    std::for_each(m_defsok.begin(), m_defsok.end(), my_delete<IndexSpec *>);
+    for (auto &ii : m_defspend) {
+        delete ii.second;
     }
 }
 
-void
-IndexSpec::to_key(const lcb_N1XSPEC* spec, std::string& s)
+void IndexSpec::to_key(const lcb_N1XSPEC *spec, std::string &out)
 {
     // Identity is:
     // {keyspace,name,is_primary,type}
-    s.append(spec->nspace, spec->nnspace).append(" ");
-    s.append(spec->keyspace, spec->nkeyspace).append(" ");
-    s.append(spec->name, spec->nname).append(" ");
+    out.append(spec->nspace, spec->nnspace).append(" ");
+    out.append(spec->keyspace, spec->nkeyspace).append(" ");
+    out.append(spec->name, spec->nname).append(" ");
     const char *type_s = ixtype_2_str(spec->ixtype);
     if (!type_s) {
         type_s = "<UNKNOWN>";
     }
-    s.append(type_s);
+    out.append(type_s);
 }
 
-void
-WatchIndexCtx::read_state(const lcb_RESPN1XMGMT *resp)
+void WatchIndexCtx::read_state(const lcb_RESPN1XMGMT *resp)
 {
     // We examine the indexes here to see which ones have been completed
     // Make them all into an std::map
@@ -645,22 +636,22 @@ WatchIndexCtx::read_state(const lcb_RESPN1XMGMT *resp)
         return;
     }
 
-    std::map<std::string, const lcb_N1XSPEC*> in_specs;
+    std::map<std::string, const lcb_N1XSPEC *> in_specs;
     for (size_t ii = 0; ii < resp->nspecs; ++ii) {
         std::string key;
         IndexSpec::to_key(resp->specs[ii], key);
         in_specs[key] = resp->specs[ii];
     }
 
-    std::map<std::string, IndexSpec*>::iterator it_remain = m_defspend.begin();
+    auto it_remain = m_defspend.begin();
     while (it_remain != m_defspend.end()) {
         // See if the index is 'online' yet!
-        std::map<std::string,const lcb_N1XSPEC*>::iterator res;
+        std::map<std::string, const lcb_N1XSPEC *>::iterator res;
         res = in_specs.find(it_remain->first);
         if (res == in_specs.end()) {
             lcb_log(LOGARGS(this, INFO), LOGFMT "Index [%s] not in cluster", LOGID(this), it_remain->first.c_str());
             // We can't find our own index. Someone else deleted it. Bail!
-            finish(LCB_KEY_ENOENT, resp);
+            finish(LCB_ERR_DOCUMENT_NOT_FOUND, resp);
             return;
         }
 
@@ -681,25 +672,23 @@ WatchIndexCtx::read_state(const lcb_RESPN1XMGMT *resp)
     }
 }
 
-lcb_error_t
-WatchIndexCtx::load_defs(const lcb_CMDN1XWATCH *cmd)
+lcb_STATUS WatchIndexCtx::load_defs(const lcb_CMDN1XWATCH *cmd)
 {
     for (size_t ii = 0; ii < cmd->nspec; ++ii) {
         std::string key;
-        IndexSpec *extspec = new IndexSpec(cmd->specs[ii]);
+        auto *extspec = new IndexSpec(cmd->specs[ii]);
         IndexSpec::to_key(extspec, key);
         m_defspend[key] = extspec;
     }
     if (m_defspend.empty()) {
-        return LCB_ENO_COMMANDS;
+        return LCB_ERR_NO_COMMANDS;
     }
     return LCB_SUCCESS;
 }
 
-void
-WatchIndexCtx::finish(lcb_error_t rc, const lcb_RESPN1XMGMT *resp)
+void WatchIndexCtx::finish(lcb_STATUS rc, const lcb_RESPN1XMGMT *resp)
 {
-    lcb_RESPN1XMGMT my_resp = { 0 };
+    lcb_RESPN1XMGMT my_resp{};
     my_resp.cookie = cookie;
     my_resp.rc = rc;
 
@@ -707,61 +696,59 @@ WatchIndexCtx::finish(lcb_error_t rc, const lcb_RESPN1XMGMT *resp)
         my_resp.inner = resp->inner;
     }
 
-    lcb_N1XSPEC **speclist = reinterpret_cast<lcb_N1XSPEC**>(&m_defsok[0]);
+    auto **speclist = reinterpret_cast<lcb_N1XSPEC **>(&m_defsok[0]);
     my_resp.specs = speclist;
     my_resp.nspecs = m_defsok.size();
     callback(m_instance, LCB_CALLBACK_IXMGMT, &my_resp);
     delete this;
 }
 
-void
-WatchIndexCtx::reschedule()
+void WatchIndexCtx::reschedule()
 {
     // Next interval!
     uint64_t now = lcb_nstime();
     if (now + LCB_US2NS(m_interval) >= m_tsend) {
-        finish(LCB_ETIMEDOUT, NULL);
+        finish(LCB_ERR_TIMEOUT, nullptr);
     } else {
         lcbio_timer_rearm(m_timer, m_interval);
     }
 }
 
-static void
-cb_watch_gotlist(lcb_t, int, const lcb_RESPN1XMGMT *resp)
+static void cb_watch_gotlist(lcb_INSTANCE *, int, const lcb_RESPN1XMGMT *resp)
 {
-    WatchIndexCtx *ctx = reinterpret_cast<WatchIndexCtx*>(resp->cookie);
+    auto *ctx = reinterpret_cast<WatchIndexCtx *>(resp->cookie);
     ctx->read_state(resp);
 }
 
-lcb_error_t
-WatchIndexCtx::do_poll()
+lcb_STATUS WatchIndexCtx::do_poll()
 {
     lcb_CMDN1XMGMT cmd;
     memset(&cmd, 0, sizeof cmd);
     cmd.callback = cb_watch_gotlist;
-    lcb_log(LOGARGS(this, DEBUG), LOGFMT "Will check for index readiness of %lu indexes. %lu completed", LOGID(this), m_defspend.size(), m_defsok.size());
+    lcb_log(LOGARGS(this, DEBUG), LOGFMT "Will check for index readiness of %lu indexes. %lu completed", LOGID(this),
+            (unsigned long int)m_defspend.size(), (unsigned long int)m_defsok.size());
     return lcb_n1x_list(m_instance, this, &cmd);
 }
 
 LIBCOUCHBASE_API
-lcb_error_t
-lcb_n1x_watchbuild(lcb_t instance, const void *cookie, const lcb_CMDN1XWATCH *cmd)
+lcb_STATUS lcb_n1x_watchbuild(lcb_INSTANCE *instance, const void *cookie, const lcb_CMDN1XWATCH *cmd)
 {
-    WatchIndexCtx *ctx = new WatchIndexCtx(instance, cookie, cmd);
-    lcb_error_t rc;
-    if ((rc = ctx->load_defs(cmd)) != LCB_SUCCESS) {
+    auto *ctx = new WatchIndexCtx(instance, cookie, cmd);
+    lcb_STATUS rc = ctx->load_defs(cmd);
+    if (rc != LCB_SUCCESS) {
         delete ctx;
         return rc;
     }
-    if ((rc = ctx->do_poll()) != LCB_SUCCESS) {
+    rc = ctx->do_poll();
+    if (rc != LCB_SUCCESS) {
         delete ctx;
         return rc;
     }
     return LCB_SUCCESS;
 }
 
-void
-IndexSpec::load_json(const char *s, size_t n) {
+void IndexSpec::load_json(const char *s, size_t n)
+{
     Json::Value root;
     // Set the JSON first!
     m_buf.assign(s, n);
@@ -792,9 +779,9 @@ IndexSpec::load_json(const char *s, size_t n) {
 }
 
 // IndexSpec stuff
-IndexSpec::IndexSpec(const lcb_N1XSPEC *spec)
+IndexSpec::IndexSpec(const lcb_N1XSPEC *spec) : lcb_N1XSPEC()
 {
-    *static_cast<lcb_N1XSPEC*>(this) = *spec;
+    *static_cast<lcb_N1XSPEC *>(this) = *spec;
     if (spec->nrawjson) {
         load_json(spec->rawjson, spec->nrawjson);
         return;
@@ -810,8 +797,7 @@ IndexSpec::IndexSpec(const lcb_N1XSPEC *spec)
     load_field(&cond, spec->cond, ncond);
 }
 
-size_t
-IndexSpec::load_fields(const Json::Value& root, bool do_copy)
+size_t IndexSpec::load_fields(const Json::Value &root, bool do_copy)
 {
     size_t size = 0;
     size += load_json_field(root, "name", &name, &nname, do_copy);
@@ -823,15 +809,14 @@ IndexSpec::load_fields(const Json::Value& root, bool do_copy)
     return size;
 }
 
-size_t
-IndexSpec::load_json_field(const Json::Value& root,
-    const char *name_, const char **tgt_ptr, size_t *tgt_len, bool do_copy)
+size_t IndexSpec::load_json_field(const Json::Value &root, const char *name_, const char **tgt_ptr, size_t *tgt_len,
+                                  bool do_copy)
 {
     size_t namelen = strlen(name_);
     const Json::Value *val = root.find(name_, name_ + namelen);
     size_t n = 0;
 
-    if (val == NULL) {
+    if (val == nullptr) {
         return 0;
     }
 
@@ -849,10 +834,10 @@ IndexSpec::load_json_field(const Json::Value& root,
     }
 
     if (n) {
-        *tgt_ptr = &(m_buf.c_str()[m_buf.size()-n]);
+        *tgt_ptr = &(m_buf.c_str()[m_buf.size() - n]);
         *tgt_len = n;
     } else {
-        *tgt_ptr = NULL;
+        *tgt_ptr = nullptr;
         *tgt_len = 0;
     }
     return n;
