@@ -6,8 +6,7 @@ Future::Future()
     failed = false;
 }
 
-void
-Future::wait()
+void Future::wait()
 {
     mutex.lock();
     while (!isDone() && !failed) {
@@ -22,14 +21,12 @@ Future::~Future()
     cond.close();
 }
 
-void
-Future::startUpdate()
+void Future::startUpdate()
 {
     mutex.lock();
 }
 
-void
-Future::endUpdate()
+void Future::endUpdate()
 {
     if (shouldEnd()) {
         cond.signal();
@@ -37,8 +34,7 @@ Future::endUpdate()
     mutex.unlock();
 }
 
-bool
-Future::checkDone()
+bool Future::checkDone()
 {
     bool ret;
     if (!mutex.tryLock()) {
